# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

"""Framework-independent types."""


class MLArtifact:
    """A framework-independent machine learning artifact."""


class MLDataset(MLArtifact):
    """A framework-independent machine learning dataset."""


class MLModel(MLArtifact):
    """A framework-independent machine learning model."""



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Henry Mutegeki
#
# Credit:
# - Henry Mutegeki <henry.mutegeki@cern.ch> - CERN
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

import matplotlib.pyplot as plt
import numpy as np
import torch
import torch.backends.cudnn as cudnn
import torch.nn as nn
import torch.nn.parallel
import torch.optim as optim
import torch.utils.data
import torchvision.datasets as dset
import torchvision.transforms as transforms
import torchvision.utils as vutils

DATA_PATH = "../data"
BATCH_SIZE = 128
IMAGE_CHANNEL = 1
Z_DIM = 100
G_HIDDEN = 64
X_DIM = 64
D_HIDDEN = 64
EPOCH_NUM = 10
REAL_LABEL = 1
FAKE_LABEL = 0
lr = 2e-4
seed = 1

USE_CUDA = False
CUDA = torch.cuda.is_available() and USE_CUDA
device = torch.device("cuda:0" if CUDA else "cpu")

print("PyTorch version: {}".format(torch.__version__))
if CUDA:
    print("CUDA version: {}\n".format(torch.version.cuda))
    torch.cuda.manual_seed(seed)

cudnn.benchmark = True

# Data preprocessing
dataset = dset.MNIST(
    root=DATA_PATH,
    download=False,
    transform=transforms.Compose(
        [transforms.Resize(X_DIM), transforms.ToTensor(), transforms.Normalize((0.5,), (0.5,))]
    ),
)

# Dataloader
dataloader = torch.utils.data.DataLoader(
    dataset, batch_size=BATCH_SIZE, shuffle=True, num_workers=2
)


def weights_init(m):
    classname = m.__class__.__name__
    if classname.find("Conv") != -1:
        m.weight.data.normal_(0.0, 0.02)
    elif classname.find("BatchNorm") != -1:
        m.weight.data.normal_(1.0, 0.02)
        m.bias.data.fill_(0)


class Generator(nn.Module):
    def __init__(self):
        super(Generator, self).__init__()
        self.main = nn.Sequential(
            # input layer
            nn.ConvTranspose2d(Z_DIM, G_HIDDEN * 8, 4, 1, 0, bias=False),
            nn.BatchNorm2d(G_HIDDEN * 8),
            nn.ReLU(True),
            # 1st hidden layer
            nn.ConvTranspose2d(G_HIDDEN * 8, G_HIDDEN * 4, 4, 2, 1, bias=False),
            nn.BatchNorm2d(G_HIDDEN * 4),
            nn.ReLU(True),
            # 2nd hidden layer
            nn.ConvTranspose2d(G_HIDDEN * 4, G_HIDDEN * 2, 4, 2, 1, bias=False),
            nn.BatchNorm2d(G_HIDDEN * 2),
            nn.ReLU(True),
            # 3rd hidden layer
            nn.ConvTranspose2d(G_HIDDEN * 2, G_HIDDEN, 4, 2, 1, bias=False),
            nn.BatchNorm2d(G_HIDDEN),
            nn.ReLU(True),
            # output layer
            nn.ConvTranspose2d(G_HIDDEN, IMAGE_CHANNEL, 4, 2, 1, bias=False),
            nn.Tanh(),
        )

    def forward(self, input):
        return self.main(input)


class Discriminator(nn.Module):
    def __init__(self):
        super(Discriminator, self).__init__()
        self.main = nn.Sequential(
            # 1st layer
            nn.Conv2d(IMAGE_CHANNEL, D_HIDDEN, 4, 2, 1, bias=False),
            nn.LeakyReLU(0.2, inplace=True),
            # 2nd layer
            nn.Conv2d(D_HIDDEN, D_HIDDEN * 2, 4, 2, 1, bias=False),
            nn.BatchNorm2d(D_HIDDEN * 2),
            nn.LeakyReLU(0.2, inplace=True),
            # 3rd layer
            nn.Conv2d(D_HIDDEN * 2, D_HIDDEN * 4, 4, 2, 1, bias=False),
            nn.BatchNorm2d(D_HIDDEN * 4),
            nn.LeakyReLU(0.2, inplace=True),
            # 4th layer
            nn.Conv2d(D_HIDDEN * 4, D_HIDDEN * 8, 4, 2, 1, bias=False),
            nn.BatchNorm2d(D_HIDDEN * 8),
            nn.LeakyReLU(0.2, inplace=True),
            # output layer
            nn.Conv2d(D_HIDDEN * 8, 1, 4, 1, 0, bias=False),
            nn.Sigmoid(),
        )

    def forward(self, input):
        return self.main(input).view(-1, 1).squeeze(1)


# Create the generator
netG = Generator().to(device)
netG.apply(weights_init)
print(netG)


# Create the discriminator
netD = Discriminator().to(device)
netD.apply(weights_init)
print(netD)


# Initialize BCELoss function
criterion = nn.BCELoss()

# Create batch of latent vectors to visualize the progression of the generator
viz_noise = torch.randn(BATCH_SIZE, Z_DIM, 1, 1, device=device)

# Setup Adam optimizers for both G and D
optimizerD = optim.Adam(netD.parameters(), lr=lr, betas=(0.5, 0.999))
optimizerG = optim.Adam(netG.parameters(), lr=lr, betas=(0.5, 0.999))


# Training Loop
def train_GAN_model(EPOCH_NUM, netD, netG, optimizerG, optimizerD, dataloader, criterion):
    img_list = []
    G_losses = []
    D_losses = []
    iters = 0
    for epoch in range(EPOCH_NUM):
        for i, data in enumerate(dataloader, 0):
            # (1) Update the discriminator with real data
            netD.zero_grad()
            # Format batch
            real_cpu = data[0].to(device)
            b_size = real_cpu.size(0)
            label = torch.full((b_size,), REAL_LABEL, dtype=torch.float, device=device)
            # Forward pass real batch through D
            output = netD(real_cpu).view(-1)
            # Calculate loss on all-real batch
            errD_real = criterion(output, label)
            # Calculate gradients for D in backward pass
            errD_real.backward()
            D_x = output.mean().item()

            # (2) Update the discriminator with fake data
            # Generate batch of latent vectors
            noise = torch.randn(b_size, Z_DIM, 1, 1, device=device)
            # Generate fake image batch with G
            fake = netG(noise)
            label.fill_(FAKE_LABEL)
            # Classify all fake batch with D
            output = netD(fake.detach()).view(-1)
            # Calculate D's loss on the all-fake batch
            errD_fake = criterion(output, label)
            # Calculate the gradients for this batch, accumulated (summed)
            # with previous gradients
            errD_fake.backward()
            D_G_z1 = output.mean().item()
            # Compute error of D as sum over the fake and the real batches
            errD = errD_real + errD_fake
            # Update D
            optimizerD.step()

            # (3) Update the generator with fake data
            netG.zero_grad()
            label.fill_(REAL_LABEL)  # fake labels are real for generator cost
            # Since we just updated D, perform another forward pass of
            # all-fake batch through D
            output = netD(fake).view(-1)
            # Calculate G's loss based on this output
            errG = criterion(output, label)
            # Calculate gradients for G
            errG.backward()
            D_G_z2 = output.mean().item()
            # Update G
            optimizerG.step()

            # Output training stats
            if i % 50 == 0:
                print(
                    (
                        "[ % d/%d][%d/%d]\tLoss_D: % .4f\tLoss_G:"
                        " % .4f\tD(x): % .4f\tD(G(z)): % .4f / %.4f"
                    )
                    % (
                        epoch,
                        EPOCH_NUM,
                        i,
                        len(dataloader),
                        errD.item(),
                        errG.item(),
                        D_x,
                        D_G_z1,
                        D_G_z2,
                    )
                )

            # Save Losses for plotting later
            G_losses.append(errG.item())
            D_losses.append(errD.item())

            # Check how the generator is doing
            if (iters % 500 == 0) or ((epoch == EPOCH_NUM - 1) and (i == len(dataloader) - 1)):
                with torch.no_grad():
                    fake = netG(viz_noise).detach().cpu()
                img_list.append(vutils.make_grid(fake, padding=2, normalize=True))

            iters += 1

    plt.figure(figsize=(10, 5))
    plt.title("Generator and Discriminator Loss During Training")
    plt.plot(G_losses, label="G")
    plt.plot(D_losses, label="D")
    plt.xlabel("iterations")
    plt.ylabel("Loss")
    plt.legend()
    plt.show()
    # plt.savefig('simpleGANlearning_curve.png')
    # Grab a batch of real images from the dataloader
    real_batch = next(iter(dataloader))
    # Plot the real images
    plt.figure(figsize=(15, 15))
    plt.subplot(1, 2, 1)
    plt.axis("off")
    plt.title("Real Images")
    plt.imshow(
        np.transpose(
            vutils.make_grid(real_batch[0].to(device)[:64], padding=5, normalize=True).cpu(),
            (1, 2, 0),
        )
    )
    # plt.savefig('simpleGANreal_image.png')
    # Plot the fake images from the last epoch
    plt.subplot(1, 2, 2)
    plt.axis("off")
    plt.title("Fake Images")
    plt.imshow(np.transpose(img_list[-1], (1, 2, 0)))
    plt.show()
    # plt.savefig('simpleGANfake_image.png')


train_GAN_model(EPOCH_NUM, netD, netG, optimizerG, optimizerD, dataloader, criterion)



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Jarl Sondre Sæther
#
# Credit:
# - Jarl Sondre Sæther <jarl.sondre.saether@cern.ch> - CERN
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

import uuid
from itertools import cycle
from pathlib import Path
from re import Match, Pattern, compile
from typing import Optional, Union

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns


def convert_matching_files_to_dataframe(
    log_dir: Path, pattern: Optional[str], expected_columns: Optional[set] = None
) -> pd.DataFrame:
    """Reads and combines all files in a folder that matches the given regex pattern
    into a single DataFrame. The files must be formatted as csv files. If pattern is
    None, we assume a match on all files.

    Raises:
        ValueError: If not all expected columns are found in the stored DataFrame.
        ValueError: If no matching files are found in the given logging directory.
    """
    re_pattern: Optional[Pattern] = None
    if pattern is not None:
        re_pattern = compile(pattern)

    if expected_columns is None:
        expected_columns = set()

    dataframes = []
    for entry in log_dir.iterdir():
        match: Union[bool, Match] = True
        if re_pattern is not None:
            match = re_pattern.search(str(entry))

        if not match:
            continue

        df = pd.read_csv(entry)
        if not expected_columns.issubset(df.columns):
            missing_columns = expected_columns - set(df.columns)
            raise ValueError(
                f"Invalid data format! File at '{str(entry)}' doesn't contain all"
                f" necessary columns. \nMissing columns: {missing_columns}"
            )

        dataframes.append(df)

    if len(dataframes) == 0:
        if pattern is None:
            error_message = f"Unable to find any files in {log_dir.resolve()}!"
        else:
            error_message = (
                f"No files matched pattern, '{pattern}', in log_dir, " f"{log_dir.resolve()}!"
            )
        raise ValueError(error_message)

    return pd.concat(dataframes)


def create_absolute_plot(avg_epoch_time_df: pd.DataFrame) -> None:
    """Creates a plot showing the absolute training times for the different
    distributed strategies and different number of GPUs.
    """
    sns.set_theme()
    fig, ax = plt.subplots()

    marker_cycle = cycle("ov^s*dXpD.+12348")

    unique_nodes = list(avg_epoch_time_df["nodes"].unique())
    unique_names = avg_epoch_time_df["name"].unique()
    for name in unique_names:
        # color, marker = next(color_marker_combinations)
        marker = next(marker_cycle)
        data = avg_epoch_time_df[avg_epoch_time_df["name"] == name]

        ax.plot(
            data["nodes"],
            data["time"],
            marker=marker,
            label=name,
            linestyle="-",
            markersize=6,
        )

    ax.set_yscale("log")
    ax.set_xscale("log")

    ax.set_xticks(unique_nodes)

    ax.set_xlabel("Number of Nodes")
    ax.set_ylabel("Average Time (s)")
    ax.set_title("Average Time vs Number of Nodes")

    ax.legend(title="Method")
    ax.grid(True)

    output_path = Path("plots/absolute_scalability_plot.png")
    output_path.parent.mkdir(parents=True, exist_ok=True)
    plt.tight_layout()
    plt.savefig(output_path)
    print(f"Saving absolute plot to '{output_path.resolve()}'.")
    sns.reset_orig()


def create_relative_plot(avg_epoch_time_df: pd.DataFrame, gpus_per_node: int = 4):
    """Creates a plot showing the relative training times for the different
    distributed strategies and different number of GPUs. In particular, it shows the
    speedup when adding more GPUs, compared to the baseline of using a single node.
    """
    sns.set_theme()

    fig, ax = plt.subplots(figsize=(6, 4))
    # fig.suptitle(plot_title)

    ax.set_yscale("log")
    ax.set_xscale("log")

    marker_cycle = cycle("ov^s*dXpD.+12348")
    avg_epoch_time_df["num_gpus"] = avg_epoch_time_df["nodes"] * gpus_per_node
    avg_epoch_time_df["linear_speedup"] = avg_epoch_time_df["nodes"].astype(float)

    # Plotting the speedup for each strategy
    strategy_names = sorted(avg_epoch_time_df["name"].unique())
    for strategy in strategy_names:
        strategy_data = avg_epoch_time_df[avg_epoch_time_df.name == strategy]

        base_time = strategy_data["time"].iloc[0]
        speedup = base_time / strategy_data["time"]
        num_gpus = strategy_data["num_gpus"]

        marker = next(marker_cycle)
        ax.plot(num_gpus, speedup, marker=marker, lw=1.0, label=strategy, alpha=0.7)

    # Plotting the linear line
    num_gpus = np.array(avg_epoch_time_df["num_gpus"].unique())
    linear_speedup = np.array(avg_epoch_time_df["linear_speedup"].unique())
    ax.plot(num_gpus, linear_speedup, ls="dashed", lw=1.0, c="k", label="linear speedup")

    ax.legend(ncol=1)
    ax.set_xticks(num_gpus)
    ax.get_xaxis().set_major_formatter(matplotlib.ticker.ScalarFormatter())
    ax.set_ylabel("Speedup")
    ax.set_xlabel("Number of GPUs (4 per node)")
    ax.grid(True)

    # Sorted legend
    handles, labels = ax.get_legend_handles_labels()
    sorted_handles_labels = sorted(zip(handles, labels), key=lambda x: x[1])
    sorted_handles, sorted_labels = zip(*sorted_handles_labels)
    plt.legend(sorted_handles, sorted_labels)

    plot_path = Path("plots/relative_scalability_plot.png")
    plt.tight_layout()
    plt.savefig(plot_path, bbox_inches="tight", format="png", dpi=300)
    print(f"Saving relative plot to '{plot_path.resolve()}'.")

    sns.reset_orig()


def backup_scalability_metrics(
    metric_df: pd.DataFrame,
    experiment_name: Optional[str],
    run_name: Optional[str],
    backup_dir: str,
    filename: str,
) -> None:
    """Stores the data in the given dataframe as a .csv file in its own folder for the
    experiment name and its own subfolder for the run_name. If these are not provided,
    then they will be generated randomly using uuid4.
    """
    if experiment_name is None:
        random_id = str(uuid.uuid4())
        experiment_name = "exp_" + random_id[:6]
    if run_name is None:
        random_id = str(uuid.uuid4())
        run_name = "run_" + random_id[:6]

    backup_path = Path(backup_dir) / experiment_name / run_name / filename
    backup_path.parent.mkdir(parents=True, exist_ok=True)
    metric_df.to_csv(backup_path, index=False)
    print(f"Storing backup file at '{backup_path.resolve()}'.")



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Anna Lappe
#
# Credit:
# - Anna Lappe <anna.elisa.lappe@cern.ch> - CERN
# --------------------------------------------------------------------------------------

from typing import Dict

from ray.tune.schedulers import (
    AsyncHyperBandScheduler,
    HyperBandForBOHB,
    HyperBandScheduler,
    PopulationBasedTraining,
)
from ray.tune.schedulers.pb2 import PB2  # Population Based Bandits
from ray.tune.search.ax import AxSearch
from ray.tune.search.bayesopt import BayesOptSearch
from ray.tune.search.bohb import TuneBOHB
from ray.tune.search.hebo import HEBOSearch
from ray.tune.search.hyperopt import HyperOptSearch
from ray.tune.search.nevergrad import NevergradSearch
from ray.tune.search.optuna import OptunaSearch
from ray.tune.search.zoopt import ZOOptSearch


def get_raytune_search_alg(
    tune_config: Dict,
) -> (
    TuneBOHB
    | BayesOptSearch
    | HyperOptSearch
    | AxSearch
    | HEBOSearch
    | NevergradSearch
    | OptunaSearch
    | ZOOptSearch
    | None
):
    """Get the appropriate Ray Tune search algorithm based on the provided configuration.

    Args:
        tune_config (Dict): Configuration dictionary specifying the search algorithm,
            metric, mode, and, depending on the search algorithm, other parameters.
        seeds (bool, optional): Whether to use a fixed seed for reproducibility for some
            search algorithms that take a seed. Defaults to False.

    Returns:
        An instance of the chosen Ray Tune search algorithm or None if no search algorithm is
        used or if the search algorithm does not match any of the supported options.
    """
    scheduler_name = tune_config.get("scheduler", {}).get("name", "")

    match scheduler_name.lower():
        case "pbt" | "pb2":
            print(
                f"INFO: Using scheduler {scheduler_name} "
                "is not compatible with Ray Tune search algorithms."
            )
            print(f"Using the Ray Tune {scheduler_name} scheduler without search algorithm")
            return None

        case "bohb":
            print(
                "INFO: Using TuneBOHB search algorithm since it is required for BOHB "
                "scheduler."
            )
            return TuneBOHB()

    search_alg = tune_config.pop("search_alg", {})
    search_alg_name = search_alg.pop("name", "")

    try:
        match search_alg_name.lower():
            case "ax":
                return AxSearch()
            case "bayesopt":
                return BayesOptSearch(**search_alg)
            case "hyperopt":
                return HyperOptSearch(**search_alg)
            case "bohb":
                return TuneBOHB(**search_alg)
            case "hepo":
                return HEBOSearch(**search_alg)
            case "nevergrad":
                return NevergradSearch(**search_alg)
            case "optuna":
                return OptunaSearch(**search_alg)
            case "zoo":
                return ZOOptSearch(**search_alg)
            case _:
                print(
                    "INFO: No search algorithm detected. Using Ray Tune BasicVariantGenerator."
                )
                return None
    except AttributeError as e:
        print(
            "Invalid search algorithm configuration passed. Please make sure that the search "
            "algorithm you are using has the correct attributes. You can read more about the "
            "different search algorithms supported by Ray Tune at "
            "https://docs.ray.io/en/latest/tune/api/suggestion.html. "
        )
        print(e)


def get_raytune_scheduler(
    tune_config: Dict,
) -> (
    AsyncHyperBandScheduler
    | HyperBandScheduler
    | HyperBandForBOHB
    | PopulationBasedTraining
    | PB2
    | None
):
    """Get the appropriate Ray Tune scheduler based on the provided configuration.

    Args:
        tune_config (Dict): Configuration dictionary specifying the scheduler type,
            metric, mode, and, depending on the scheduler, other parameters.

    Returns:
        An instance of the chosen Ray Tune scheduler or None if no scheduler is used
        or if the scheduler does not match any of the supported options.
    """

    scheduler = tune_config.pop("scheduler", {})
    scheduler_name = scheduler.pop("name", "")

    try:
        match scheduler_name.lower():
            case "asha":
                return AsyncHyperBandScheduler(**scheduler)
            case "hyperband":
                return HyperBandScheduler(**scheduler)
            case "bohb":
                return HyperBandForBOHB(**scheduler)
            case "pbt":
                return PopulationBasedTraining(**scheduler)
            case "pb2":
                return PB2(**scheduler)
            case _:
                print(
                    "INFO: No search algorithm detected. Using default Ray Tune FIFOScheduler."
                )
                return None
    except AttributeError as e:
        print(
            "Invalid scheduler configuration passed. Please make sure that the scheduler "
            "you are using has the correct attributes. You can read more about the "
            "different schedulers supported by Ray Tune at "
            "https://docs.ray.io/en/latest/tune/api/schedulers.html."
        )
        print(e)



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

from typing import Optional

import lightning as L
from torch.utils.data import DataLoader, random_split
from torchvision import transforms
from torchvision.datasets import MNIST

from itwinai.components import DataGetter, monitor_exec


class LightningMNISTDownloader(DataGetter):
    def __init__(self, data_path: str, name: Optional[str] = None) -> None:
        super().__init__(name)
        self.save_parameters(**self.locals2params(locals()))
        self.data_path = data_path
        self._downloader = MNISTDataModule(
            data_path=self.data_path,
            download=True,
            # Mock other args...
            batch_size=1,
            train_prop=0.5,
        )

    @monitor_exec
    def execute(self) -> None:
        # Simulate dataset creation to force data download
        self._downloader.setup(stage="fit")
        self._downloader.setup(stage="test")
        self._downloader.setup(stage="predict")


class MNISTDataModule(L.LightningDataModule):
    def __init__(
        self, data_path: str, batch_size: int, train_prop: float, download: bool = True
    ) -> None:
        super().__init__()
        self.data_path = data_path
        self.download = download
        self.batch_size = batch_size
        self.train_prop = train_prop
        self.transform = transforms.Compose(
            [
                transforms.ToTensor(),
                transforms.Normalize((0.1307,), (0.3081,)),
            ]
        )

    def setup(self, stage=None):
        if stage == "fit":
            mnist_full = MNIST(
                self.data_path, train=True, download=self.download, transform=self.transform
            )
            n_train_samples = int(self.train_prop * len(mnist_full))
            n_val_samples = len(mnist_full) - n_train_samples
            self.mnist_train, self.mnist_val = random_split(
                mnist_full, [n_train_samples, n_val_samples]
            )

        if stage == "test":
            self.mnist_test = MNIST(
                self.data_path, train=False, download=self.download, transform=self.transform
            )

        if stage == "predict":
            self.mnist_predict = MNIST(
                self.data_path, train=False, download=self.download, transform=self.transform
            )

    def train_dataloader(self):
        return DataLoader(self.mnist_train, batch_size=self.batch_size, num_workers=4)

    def val_dataloader(self):
        return DataLoader(self.mnist_val, batch_size=self.batch_size, num_workers=4)

    def test_dataloader(self):
        return DataLoader(self.mnist_test, batch_size=self.batch_size, num_workers=4)

    def predict_dataloader(self):
        return DataLoader(self.mnist_predict, batch_size=self.batch_size, num_workers=4)



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# -------------------------------------------------------------------------------------

from unittest.mock import MagicMock, patch

import numpy as np


def test_console_logger_log(console_logger):
    console_logger.create_logger_context()
    with patch("builtins.print") as mocked_print:
        console_logger.log("test_value", "test_identifier")
        mocked_print.assert_called_with("ConsoleLogger: test_identifier = test_value")
    console_logger.destroy_logger_context()


def test_mlflow_logger_log(mlflow_logger):
    with patch("mlflow.log_metric") as mock_log_metric:
        mlflow_logger.create_logger_context()
        mlflow_logger.log(0.5, "test_metric", kind="metric", step=1)
        mock_log_metric.assert_called_once_with(key="test_metric", value=0.5, step=1)
        mlflow_logger.destroy_logger_context()


def test_wandb_logger_log(wandb_logger):
    with patch("wandb.init") as mock_init, patch("wandb.log") as mock_log:
        mock_init.return_value = MagicMock()
        wandb_logger.create_logger_context()
        wandb_logger.log(0.5, "test_metric", kind="metric")
        mock_log.assert_called_once_with({"test_metric": 0.5}, commit=True)


def test_tensorboard_logger_log_tf(tensorboard_logger_tf):
    import tensorflow as tf

    with (
        patch("tensorflow.summary.scalar") as mock_scalar,
        patch("tensorflow.summary.image") as mock_image,
        patch("tensorflow.summary.text") as mock_text,
    ):
        tensorboard_logger_tf.create_logger_context()

        # Log a scalar
        tensorboard_logger_tf.log(0.5, "test_scalar", kind="metric", step=1)
        mock_scalar.assert_called_once_with("test_scalar", 0.5, step=1)

        # Log an image
        image = tf.zeros([10, 10, 3])
        tensorboard_logger_tf.log(image, "test_image", kind="image", step=1)
        mock_image.assert_called_once()
        assert np.allclose(mock_image.call_args[0][1].numpy(), image.numpy())

        # Log text
        tensorboard_logger_tf.log("test text", "test_text", kind="text", step=1)
        mock_text.assert_called_once_with("test_text", "test text", step=1)

        tensorboard_logger_tf.destroy_logger_context()


def test_tensorboard_logger_log_torch(tensorboard_logger_torch):
    import torch

    step = 1
    with (
        patch("torch.utils.tensorboard.SummaryWriter.add_scalar") as mock_scalar,
        patch("torch.utils.tensorboard.SummaryWriter.add_image") as mock_image,
        patch("torch.utils.tensorboard.SummaryWriter.add_text") as mock_text,
    ):
        tensorboard_logger_torch.create_logger_context()

        # Log a scalar
        tensorboard_logger_torch.log(0.5, "test_scalar", kind="metric", step=step)
        mock_scalar.assert_called_once_with("test_scalar", 0.5, global_step=step)

        # Log an image
        image = torch.zeros([3, 10, 10])
        tensorboard_logger_torch.log(image, "test_image", kind="image", step=step)
        mock_image.assert_called_once_with("test_image", image, global_step=step)

        # Log text
        tensorboard_logger_torch.log("test text", "test_text", kind="text", step=step)
        mock_text.assert_called_once_with("test_text", "test text", global_step=step)

        tensorboard_logger_torch.destroy_logger_context()


def test_loggers_collection_log(loggers_collection):
    with (
        patch("builtins.print") as mocked_print,
        patch("mlflow.log_metric") as mock_log_metric,
        patch("wandb.init") as mock_wandb_init,
        patch("wandb.log") as mock_wandb_log,
    ):  # , \
        # patch('shutil.copyfile') as mock_shutil_copyfile, \
        # patch('pickle.dump') as mock_pickle_dump, \
        # patch('torch.save') as mock_torch_save:

        mock_wandb_init.return_value = MagicMock()

        loggers_collection.create_logger_context()
        loggers_collection.log(0.5, "test_metric", kind="metric", step=1)

        mocked_print.assert_called_with("ConsoleLogger: test_metric = 0.5")
        mock_log_metric.assert_called_once_with(key="test_metric", value=0.5, step=1)
        mock_wandb_log.assert_called_once_with({"test_metric": 0.5}, commit=True)

        loggers_collection.destroy_logger_context()


# @patch('prov4ml.start_run')
# def test_create_logger_context(mock_start_run, prov4ml_logger):
#     prov4ml_logger.create_logger_context()
#     mock_start_run.assert_called_once_with(
#         prov_user_namespace="www.example.org",
#         experiment_name="experiment_name",
#         provenance_save_dir="prov",
#         save_after_n_logs=100,
#         collect_all_processes=False
#     )


# @patch('prov4ml.end_run')
# def test_destroy_logger_context(mock_end_run, prov4ml_logger):
#     prov4ml_logger.destroy_logger_context()
#     mock_end_run.assert_called_once_with(create_graph=True, create_svg=True)


# @patch('prov4ml.log_metric')
# def test_log_metric(mock_log_metric, prov4ml_logger):
#     item = 1.0
#     identifier = "test_metric"
#     context = MagicMock(spec=Context)
#     prov4ml_logger.log(
#         item, identifier, kind='metric', step=0, context=context)
#     mock_log_metric.assert_called_once_with(identifier, item, context, step=0)


# @patch('prov4ml.log_flops_per_batch')
# def test_log_flops_per_batch(mock_log_flops_per_batch, prov4ml_logger):
#     item = (MagicMock(), MagicMock())
#     identifier = "test_flops"
#     context = MagicMock(spec=Context)
#     prov4ml_logger.log(
#         item, identifier, kind='flops_pb', step=0,
#         context=context)
#     mock_log_flops_per_batch.assert_called_once_with(
#         identifier, model=item[0], batch=item[1], context=context, step=0)



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

import os
from pathlib import Path

import pytest


@pytest.fixture
def torch_env() -> str:
    """If TORCH_ENV env variable is defined, it overrides the default
    torch virtual environment name. Otherwise, fall back
    to './.venv-pytorch'.

    Returns absolute path to torch virtual environment.
    """
    env_path = Path(os.environ.get("TORCH_ENV", "./.venv-pytorch"))
    return str(env_path.resolve())


@pytest.fixture
def tf_env() -> str:
    """If TF_ENV env variable is defined, it overrides the default
    torch virtual environment name. Otherwise, fall back
    to './.venv-tf'.

    Returns absolute path to torch virtual environment.
    """
    env_path = Path(os.environ.get("TF_ENV", "./.venv-tf"))
    return str(env_path.resolve())



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# - Jarl Sondre Sæther <jarl.sondre.saether@cern.ch> - CERN
# --------------------------------------------------------------------------------------


"""Scaling test of torch Distributed Data Parallel on Imagenet using Resnet."""

import os
from pathlib import Path
from timeit import default_timer as timer

import torch
import torch.distributed as dist
import torch.nn as nn
import torchvision
from torch.utils.data import DataLoader
from torch.utils.data.distributed import DistributedSampler
from utils import get_parser, imagenet_dataset, train_epoch

from itwinai.loggers import EpochTimeTracker
from itwinai.torch.reproducibility import seed_worker, set_seed


def main():
    parser = get_parser()
    args = parser.parse_args()

    use_cuda = not args.no_cuda and torch.cuda.is_available()
    is_distributed = use_cuda and torch.cuda.device_count() > 0
    torch_seed = set_seed(args.rnd_seed, deterministic_cudnn=False)
    shuffle: bool = args.shuff and args.rnd_seed is None
    persistent_workers = args.nworker > 1
    epoch_time_save_dir = Path(args.epoch_time_directory)

    train_dataset = imagenet_dataset(args.data_dir, subset_size=args.subset_size)
    train_sampler = None

    if is_distributed:
        # Initializing the distribution backend
        dist.init_process_group(backend=args.backend)

        local_world_size = torch.cuda.device_count()
        global_rank = dist.get_rank()
        local_rank = dist.get_rank() % local_world_size

        # Creating dataset and dataloader
        pin_memory = True
        train_sampler = DistributedSampler(train_dataset, shuffle=shuffle)
    else:
        local_world_size = 1
        global_rank = 0
        local_rank = 0
        pin_memory = False

    train_loader = DataLoader(
        train_dataset,
        batch_size=args.batch_size,
        sampler=train_sampler,
        num_workers=args.nworker,
        pin_memory=pin_memory,
        persistent_workers=persistent_workers,
        prefetch_factor=args.prefetch,
        generator=torch_seed,
        worker_init_fn=seed_worker,
    )
    device = torch.device(f"cuda:{local_rank}" if use_cuda else "cpu")

    model = torchvision.models.resnet152()
    model.to(device)

    # Distributing the model to the workers
    if is_distributed:
        model = nn.parallel.DistributedDataParallel(
            model, device_ids=[device], output_device=device
        )

    optimizer = torch.optim.SGD(model.parameters(), lr=args.lr, momentum=args.momentum)

    if global_rank == 0:
        num_nodes = os.environ.get("SLURM_NNODES", 1)
        save_path = epoch_time_save_dir / f"epochtime_ddp-bl_{num_nodes}.csv"
        epoch_time_tracker = EpochTimeTracker(
            strategy_name="ddp-bl",
            save_path=save_path,
            num_nodes=int(num_nodes),
        )

    start_time = timer()
    for epoch_idx in range(1, args.epochs + 1):
        epoch_start_time = timer()

        if is_distributed:
            train_sampler.set_epoch(epoch_idx)

        train_epoch(
            model=model, device=device, train_loader=train_loader, optimizer=optimizer
        )

        if global_rank == 0:
            epoch_elapsed_time = timer() - epoch_start_time
            epoch_time_tracker.add_epoch_time(epoch_idx, epoch_elapsed_time)
            print(f"[{epoch_idx}/{args.epochs}] - time: {epoch_elapsed_time:.2f}s")

    if global_rank == 0:
        total_time = timer() - start_time
        print(f"Training finished - took {total_time:.2f}s")
        epoch_time_tracker.save()

    # Clean-up
    if is_distributed:
        dist.barrier()
        dist.destroy_process_group()


if __name__ == "__main__":
    main()



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

import abc
import builtins as __builtin__
import functools
import os
import sys
from typing import Any, Callable

from pydantic import BaseModel


class DistributedStrategy(abc.ABC):
    """Abstract class to define the distributed backend methods."""


class ClusterEnvironment(BaseModel):
    """Stores information about distributed environment."""

    #: Global rank of current worker, in a distributed environment.
    #: ``global_rank==0`` identifies the main worker.
    #: Defaults to 0.
    global_rank: int = 0
    #: Local rank of current worker, in a distributed environment.
    #: Defaults to 0.
    local_rank: int = 0
    #: Total number of workers in a distributed environment.
    #: Defaults to 1.
    global_world_size: int = 1
    #: Number of workers on the same node in a distributed environment.
    #: Defaults to 1.
    local_world_size: int = 1


def detect_distributed_environment() -> ClusterEnvironment:
    """Detects distributed environment, extracting information like
    global ans local ranks, and world size.
    """
    if os.getenv("TORCHELASTIC_RUN_ID") is not None:
        # Torch elastic environment
        # https://pytorch.org/docs/stable/elastic/run.html#environment-variables
        return ClusterEnvironment(
            global_rank=os.getenv("RANK"),
            local_rank=os.getenv("LOCAL_RANK"),
            local_world_size=os.getenv("LOCAL_WORLD_SIZE"),
            global_world_size=os.getenv("WORLD_SIZE"),
        )
    elif os.getenv("OMPI_COMM_WORLD_SIZE") is not None:
        # Open MPI environment
        # https://docs.open-mpi.org/en/v5.0.x/tuning-apps/environment-var.html
        return ClusterEnvironment(
            global_rank=os.getenv("OMPI_COMM_WORLD_RANK"),
            local_rank=os.getenv("OMPI_COMM_WORLD_LOCAL_RANK"),
            local_world_size=os.getenv("OMPI_COMM_WORLD_LOCAL_SIZE"),
            global_world_size=os.getenv("OMPI_COMM_WORLD_SIZE"),
        )
    elif os.getenv("SLURM_JOB_ID") is not None:
        print(
            "WARNING: detected SLURM environment, but "
            "unable to determine ranks and world sizes!"
        )
        return ClusterEnvironment()
    else:
        return ClusterEnvironment()


#: Save original builtin print before patching it in distributed environments
builtin_print = __builtin__.print


def distributed_patch_print(is_main: bool) -> Callable:
    """Disable ``print()`` when not in main worker.

    Args:
        is_main (bool): whether it is called from main worker.

    Returns:
        Callable: patched ``print()``.
    """

    def patched_print(*args, **kwself):
        """Print is disables on workers different from
        the main one, unless the print is called with
        ``force=True`` argument.
        """
        force = kwself.pop("force", False)
        if is_main or force:
            builtin_print(*args, **kwself)

    return patched_print


def suppress_workers_print(func: Callable) -> Callable:
    """Decorator to suppress ``print()`` calls in workers having global rank
    different from 0. To force printing on all workers you need to use
    ``print(..., force=True)``.
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs) -> Any:
        # Disable print in workers different from the main one,
        # when in distributed environments.
        dist_grank = detect_distributed_environment().global_rank
        patched_print = distributed_patch_print(is_main=dist_grank == 0)
        previous_print_backup = __builtin__.print
        __builtin__.print = patched_print
        try:
            result = func(*args, **kwargs)
        except Exception as exc:
            # Reset print to builtin before raising the exception.
            __builtin__.print = previous_print_backup
            raise exc
        # Reset print to builtin
        __builtin__.print = previous_print_backup
        return result

    return wrapper


def suppress_workers_output(func):
    """Decorator to suppress ``stadout`` and ``stderr`` in workers having global rank
    different from 0.
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        # Save the original stdout and stderr
        original_stdout = sys.stdout
        original_stderr = sys.stderr

        # Get global rank
        dist_grank = detect_distributed_environment().global_rank
        try:
            if dist_grank == 0:
                # If on main worker
                return func(*args, **kwargs)

            # If not on main worker, redirect stdout and stderr to devnull
            with open(os.devnull, "w") as devnull:
                sys.stdout = devnull
                sys.stderr = devnull
                # Execute the wrapped function
                return func(*args, **kwargs)
        finally:
            # Restore original stdout and stderr
            sys.stdout = original_stdout
            sys.stderr = original_stderr

    return wrapper



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Jarl Sondre Sæther
#
# Credit:
# - Jarl Sondre Sæther <jarl.sondre.saether@cern.ch> - CERN
# --------------------------------------------------------------------------------------

import argparse
from pathlib import Path

import h5py
import numpy as np


def append_to_hdf5_dataset(
    file_path: Path,
    dataset_name: str,
    array: np.ndarray,
    expected_datapoint_shape: tuple,
):
    if tuple(array.shape[1:]) != expected_datapoint_shape:
        actual_shape_str = ", ".join(str(s) for s in array.shape)
        expected_shape_str = ", ".join(str(s) for s in expected_datapoint_shape)
        raise ValueError(
            f"'array' has an incorrect shape: ({actual_shape_str}). "
            f"Should have been (x, {expected_shape_str})."
        )

    # print(f"Appending to file: '{str(file_path.resolve())}'.")
    with h5py.File(file_path, "a") as f:
        dset = f[dataset_name]
        dset.resize(dset.shape[0] + array.shape[0], axis=0)
        dset[-array.shape[0] :] = array


def main():
    parser = argparse.ArgumentParser(description="Virgo Dataset Generation")
    parser.add_argument(
        "--dir", type=str, help="Directory containing the HDF5 files to concatenate"
    )
    parser.add_argument(
        "--save-location",
        type=str,
        help="Location to save the resulting HDF5 file.",
        default="total_virgo_data.hdf5",
    )
    args = parser.parse_args()
    dir = Path(args.dir)
    save_location = Path(args.save_location)
    num_aux_channels = 3
    square_size = 64
    dataset_name = "virgo_dataset"

    # Creating empty HDF5 file
    datapoint_shape = (num_aux_channels + 1, square_size, square_size)
    save_location.parent.mkdir(parents=True, exist_ok=True)

    print(f"Creating/overwriting file: '{save_location.resolve()}'.")
    with h5py.File(save_location, "w") as f:
        dataset = f.create_dataset(
            dataset_name,
            shape=(0, *datapoint_shape),
            maxshape=(None, *datapoint_shape),
            dtype=np.float32,
        )
        dataset.attrs["Description"] = (
            "Synthetic time series data for the Virgo use case"
        )
        dataset.attrs["main_channel_idx"] = 0

    # NOTE: This will not necessarily iterate in same order as the suffices of the
    # file names
    files = [
        entry for entry in dir.iterdir()
        if (entry.suffix == ".hdf5" and entry.stem != "virgo_data")
    ]
    for entry in files:
        with h5py.File(entry, "r") as f:
            data = f[dataset_name][:]
        print(f"Adding {len(data)} rows from entry: {entry}")

        append_to_hdf5_dataset(
            file_path=save_location,
            dataset_name=dataset_name,
            array=data,
            expected_datapoint_shape=datapoint_shape,
        )


if __name__ == "__main__":
    main()



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# - Jarl Sondre Sæther <jarl.sondre.saether@cern.ch> - CERN
# --------------------------------------------------------------------------------------

"""Tests for MNIST use case.

Intended to be integration tests, to make sure that updates in the code base
do not break use cases' workflows.
"""

import os
import subprocess
import tempfile
from pathlib import Path

import pytest

MNIST_FOLDER = Path("use-cases", "mnist")
TORCH_PATH = MNIST_FOLDER / "torch"
LIGHTNING_PATH = MNIST_FOLDER / "torch-lightning"
TF_PATH = MNIST_FOLDER / "tensorflow"

DEFAULT_MNIST_DATASET = ".tmp"


@pytest.mark.skip(reason="structure changed")
def test_structure_mnist_torch(check_folder_structure):
    """Test MNIST folder structure for torch native trainer."""
    check_folder_structure(TORCH_PATH)


@pytest.mark.skip(reason="structure changed")
def test_structure_mnist_lightning(check_folder_structure):
    """Test MNIST folder structure for torch lightning trainer."""
    check_folder_structure(LIGHTNING_PATH)


@pytest.mark.skip(reason="structure changed")
def test_structure_mnist_tf(check_folder_structure):
    """Test MNIST folder structure for tensorflow trainer."""
    check_folder_structure(TF_PATH)


@pytest.mark.functional
def test_mnist_train_torch(torch_env, install_requirements):
    """
    Test MNIST torch native trainer by running it end-to-end.

    If MNIST_DATASET env variable is defined, it is used to
    override the default dataset download location: useful
    when it contains a local copy of the dataset, preventing
    downloading it again.
    """
    install_requirements(TORCH_PATH, torch_env)

    dataset_path = os.environ.get("MNIST_DATASET", DEFAULT_MNIST_DATASET)

    conf = (TORCH_PATH / "config.yaml").resolve()
    cmd = (
        f"{torch_env}/bin/itwinai exec-pipeline "
        f"--config {conf} --pipe-key training_pipeline "
        f"-o dataset_root={dataset_path}"
    )
    with tempfile.TemporaryDirectory() as temp_dir:
        subprocess.run(cmd.split(), check=True, cwd=temp_dir)


@pytest.mark.functional
def test_mnist_inference_torch(torch_env, install_requirements):
    """
    Test MNIST torch native inference by running it end-to-end.
    """
    install_requirements(TORCH_PATH, torch_env)

    conf = (TORCH_PATH / "config.yaml").resolve()
    exec = (TORCH_PATH / "create_inference_sample.py").resolve()

    run_inference_cmd = (
        f"{torch_env}/bin/itwinai exec-pipeline "
        f"--config {conf} --pipe-key inference_pipeline"
    )
    with tempfile.TemporaryDirectory() as temp_dir:
        # Create fake inference dataset and checkpoint
        generate_model_cmd = f"{torch_env}/bin/python {exec} " f"--root {temp_dir}"
        subprocess.run(generate_model_cmd.split(), check=True, cwd=temp_dir)

        # Running inference
        subprocess.run(run_inference_cmd.split(), check=True, cwd=temp_dir)


@pytest.mark.functional
def test_mnist_train_torch_lightning(torch_env, install_requirements):
    """
    Test MNIST torch lightning trainer by running it end-to-end.

    If MNIST_DATASET env variable is defined, it is used to
    override the default dataset download location: useful
    when it contains a local copy of the dataset, preventing
    downloading it again.
    """
    install_requirements(LIGHTNING_PATH, torch_env)

    dataset_path = os.environ.get("MNIST_DATASET", DEFAULT_MNIST_DATASET)
    conf = (LIGHTNING_PATH / "config.yaml").resolve()
    cmd = (
        f"{torch_env}/bin/itwinai exec-pipeline "
        f"--config {conf} --pipe-key training_pipeline "
        f"-o dataset_root={dataset_path}"
    )
    with tempfile.TemporaryDirectory() as temp_dir:
        subprocess.run(cmd.split(), check=True, cwd=temp_dir)


@pytest.mark.functional
def test_mnist_train_tf(tf_env, install_requirements):
    """
    Test MNIST tensorflow trainer by running it end-to-end.
    """
    install_requirements(TF_PATH, tf_env)
    conf = TF_PATH / "pipeline.yaml"
    cmd = f"{tf_env}/bin/itwinai exec-pipeline " f"--config {conf.resolve()} --pipe-key pipeline"
    with tempfile.TemporaryDirectory() as temp_dir:
        subprocess.run(cmd.split(), check=True, cwd=temp_dir)



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Jarl Sondre Sæther
#
# Credit:
# - Jarl Sondre Sæther <jarl.sondre.saether@cern.ch> - CERN
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

import functools
from pathlib import Path
from typing import Any, Callable, Iterable, Tuple

import matplotlib
import pandas as pd
from torch.profiler import ProfilerActivity, profile, schedule

from itwinai.torch.trainer import TorchTrainer

# Doing this because otherwise I get an error about X11 Forwarding which I believe
# is due to the server trying to pass the image to the client computer
matplotlib.use("Agg")


def profile_torch_trainer(method: Callable) -> Callable:
    """Decorator for execute method for components. Profiles the communication time
    vs. computation time and stores the result for future analysis.
    """

    def gather_profiling_data(key_averages: Iterable) -> pd.DataFrame:
        profiling_data = []
        for event in key_averages:
            profiling_data.append(
                {
                    "name": event.key,
                    "node_id": event.node_id,
                    "self_cuda_time_total": event.self_device_time_total,
                    "cuda_time_total": event.device_time_total,
                    "cuda_time_total_str": event.device_time_total_str,
                    "calls": event.count,
                }
            )
        return pd.DataFrame(profiling_data)

    def adjust_wait_and_warmup_epochs(
        training_epochs: int, wait_epochs: int, warmup_epochs: int
    ) -> Tuple[int, int, int]:
        """Validates if the given wait and warmup epochs are compatible and if not,
        adjusts them so they fit. The largest one is iteratively decreased until
        a compatible value is reached.

        Returns:
            int: The resulting number of epochs for doing active profiling
            int: The resulting number of wait epochs, possibly adjusted
            int: The resulting number of warmup epochs, possibly adjusted
        """
        active_epochs = training_epochs - wait_epochs - warmup_epochs
        if active_epochs > 0:
            return active_epochs, wait_epochs, warmup_epochs

        # This can probably be done with a simple math expression, but this was
        # simpler to implement and won't really cause much overhead anyway...
        while active_epochs <= 0:
            if wait_epochs > warmup_epochs:
                wait_epochs -= 1
            else:
                warmup_epochs -= 1
            active_epochs = training_epochs - wait_epochs - warmup_epochs

        if wait_epochs < 0 or warmup_epochs < 0:
            raise ValueError(
                f"Unable to adjust wait and warmup epochs to accomodate the"
                f"given number of training epochs. Was given the following values: "
                f"Training epochs: {training_epochs}, wait epochs: {wait_epochs}"
                f", warmup epochs: {warmup_epochs}"
            )
        print(
            f"Warning: adjusted the given wait and warmup epochs for the profiler - "
            f"wait epochs: {wait_epochs}, warmup epochs: {warmup_epochs}."
        )
        return active_epochs, wait_epochs, warmup_epochs

    @functools.wraps(method)
    def profiled_method(self: TorchTrainer, *args, **kwargs) -> Any:
        active_epochs, wait_epochs, warmup_epochs = adjust_wait_and_warmup_epochs(
            training_epochs=self.epochs,
            wait_epochs=self.profiling_wait_epochs,
            warmup_epochs=self.profiling_warmup_epochs,
        )
        with profile(
            activities=[ProfilerActivity.CUDA, ProfilerActivity.CPU],
            schedule=schedule(
                wait=wait_epochs,
                warmup=warmup_epochs,
                active=active_epochs,
            ),
            with_modules=True
        ) as profiler:
            self.profiler = profiler
            result = method(self, *args, **kwargs)

        strategy = self.strategy
        strategy_name = strategy.name

        global_rank = strategy.global_rank()
        num_gpus_global = strategy.global_world_size()

        # Extracting and storing the profiling data
        key_averages = profiler.key_averages()

        profiling_dataframe = gather_profiling_data(key_averages=key_averages)
        profiling_dataframe["strategy"] = strategy_name
        profiling_dataframe["num_gpus"] = num_gpus_global
        profiling_dataframe["global_rank"] = global_rank

        profiling_log_dir = Path("scalability-metrics/communication-data")
        profiling_log_dir.mkdir(parents=True, exist_ok=True)

        filename = f"{strategy_name}_{num_gpus_global}_{global_rank}.csv"
        output_path = profiling_log_dir / filename

        print(f"Writing communication profiling dataframe to '{output_path}'.")
        profiling_dataframe.to_csv(output_path)

        return result

    return profiled_method



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

"""Create a simple inference dataset sample and a checkpoint."""

import argparse
import os

import torch
from model import ThreeDGAN


def create_checkpoint(root: str = ".", ckpt_name: str = "3dgan-inference.pth"):
    ckpt_path = os.path.join(root, ckpt_name)
    net = ThreeDGAN()
    torch.save(net, ckpt_path)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--root", type=str, default=".")
    parser.add_argument("--ckpt-name", type=str, default="3dgan-inference.pth")
    args = parser.parse_args()
    create_checkpoint(**vars(args))



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Jarl Sondre Sæther
#
# Credit:
# - Jarl Sondre Sæther <jarl.sondre.saether@cern.ch> - CERN
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

import functools
import time
from multiprocessing import Manager, Process
from pathlib import Path
from typing import Any, Callable, Dict, List

import pandas as pd
import pynvml
from pynvml import nvmlDeviceGetHandleByIndex, nvmlInit

from itwinai.torch.trainer import TorchTrainer

logging_columns = [
    "sample_idx",
    "utilization",
    "power",
    "local_rank",
    "node_idx",
    "num_global_gpus",
    "strategy",
    "probing_interval",
]


def probe_gpu_utilization_loop(
    node_idx: int,
    num_local_gpus: int,
    num_global_gpus: int,
    strategy_name: str,
    log_dict: Any,
    stop_flag: Any,
    probing_interval: int = 2,
    warmup_time: int = 5,
) -> None:
    """Logs the GPU utilization across all availble GPUs on a single node. Is meant to
    be called by multiprocessing's Process and expects variables to be shared using
    a multiprocessing.Manager object. Logs utilization into `log_dict` until
    stop_flag.value is set to True.

    Args:
        node_idx: The index of the compute node that the function is called by, used
            for logging purposes.
        num_local_gpus: Number of GPUs on the current compute node.
        num_global_gpus: Number of GPUs on all nodes combined.
        strategy: Which distributed strategy is being used, e.g. "ddp" or "horovod".
        log_dict: Dictionary for storing logging data on. Should be managed by a
            multiprocessing.Manager object.
        stop_flag: Shared value telling the function when to stop logging. Should be
            managed by a multiprocessing.Manager object.
        probing_interval: How long to wait between each time a read of the GPU
            utilization is done.
        warmup_time: How long to wait before logging starts, allowing the training to
            properly start before reading.

    """

    if not set(logging_columns).issubset(set(log_dict.keys())):
        missing_columns = set(logging_columns) - set(log_dict.keys())
        raise ValueError(
            f"log_dict is missing the following columns: {missing_columns}"
        )

    nvmlInit()
    time.sleep(warmup_time)

    sample_idx = 0
    while not stop_flag.value:
        for idx in range(num_local_gpus):
            handle = nvmlDeviceGetHandleByIndex(idx)
            utilization_rates = pynvml.nvmlDeviceGetUtilizationRates(handle)

            gpu_util = utilization_rates.gpu
            power = pynvml.nvmlDeviceGetPowerUsage(handle)
            power = power / 1000  # mW -> W

            log_dict["sample_idx"].append(sample_idx)
            log_dict["utilization"].append(gpu_util)
            log_dict["power"].append(power)
            log_dict["local_rank"].append(idx)
            log_dict["node_idx"].append(node_idx)
            log_dict["num_global_gpus"].append(num_global_gpus)
            log_dict["strategy"].append(strategy_name)
            log_dict["probing_interval"].append(probing_interval)

        sample_idx += 1
        time.sleep(probing_interval)


def measure_gpu_utilization(method: Callable) -> Callable:
    """Decorator for measuring GPU utilization and storing it to a .csv file."""

    def write_logs_to_file(utilization_logs: List[Dict], output_path: Path) -> None:
        dataframes = []
        for log in utilization_logs:
            if len(log) == 0:
                continue
            dataframes.append(pd.DataFrame(log))

        log_df = pd.concat(dataframes)
        log_df.to_csv(output_path, index=False)
        print(f"Writing GPU energy dataframe to '{output_path.resolve()}'.")

    @functools.wraps(method)
    def measured_method(self: TorchTrainer, *args, **kwargs) -> Any:
        gpu_probing_interval = 1
        warmup_time = 5

        strategy = self.strategy
        strategy_name = strategy.name

        local_rank = strategy.local_rank()
        global_rank = strategy.global_rank()
        num_global_gpus = strategy.global_world_size()
        num_local_gpus = strategy.local_world_size()
        node_idx = global_rank // num_local_gpus

        gpu_monitor_process = None
        manager = None
        stop_flag = None
        data = None

        # Starting a child process once per node
        if local_rank == 0:

            # Setting up shared variables for the child process
            manager = Manager()
            data = manager.dict()
            for col in logging_columns:
                data[col] = manager.list()
            stop_flag = manager.Value("i", False)

            gpu_monitor_process = Process(
                target=probe_gpu_utilization_loop,
                kwargs={
                    "node_idx": node_idx,
                    "num_local_gpus": num_local_gpus,
                    "num_global_gpus": num_global_gpus,
                    "strategy_name": strategy_name,
                    "log_dict": data,
                    "stop_flag": stop_flag,
                    "probing_interval": gpu_probing_interval,
                    "warmup_time": warmup_time,
                },
            )
            gpu_monitor_process.start()

        local_utilization_log = {}
        try:
            result = method(self, *args, **kwargs)
        finally:
            if local_rank == 0:
                stop_flag.value = True
                grace_period = 5  # extra time to let process finish gracefully
                gpu_monitor_process.join(timeout=gpu_probing_interval + grace_period)

                # Converting the shared log to non-shared log
                local_utilization_log = {key: list(data[key]) for key in data.keys()}
                manager.shutdown()

        global_utilization_log = strategy.gather_obj(local_utilization_log, dst_rank=0)
        if strategy.is_main_worker:
            output_dir = Path("scalability-metrics/gpu-energy-data")
            output_dir.mkdir(exist_ok=True, parents=True)
            output_path = output_dir / f"{strategy_name}_{num_global_gpus}.csv"

            write_logs_to_file(global_utilization_log, output_path)

        return result

    return measured_method



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

from typing import Generator

import pytest
import torch

from itwinai.torch.distributed import (
    DeepSpeedStrategy,
    HorovodStrategy,
    TorchDDPStrategy,
    TorchDistributedStrategy,
)


@pytest.fixture(scope="package")
def ddp_strategy() -> Generator[TorchDistributedStrategy, None, None]:
    """Instantiate Torch's DistributedDataParallel strategy."""
    strategy = TorchDDPStrategy(backend="nccl" if torch.cuda.is_available() else "gloo")
    strategy.init()
    yield strategy
    strategy.clean_up()


@pytest.fixture(scope="package")
def deepspeed_strategy() -> Generator[DeepSpeedStrategy, None, None]:
    """Instantiate DeepSpeed strategy."""
    strategy = DeepSpeedStrategy(backend="nccl" if torch.cuda.is_available() else "gloo")
    strategy.init()
    yield strategy
    strategy.clean_up()


@pytest.fixture(scope="package")
def horovod_strategy() -> Generator[HorovodStrategy, None, None]:
    """Instantiate Horovod strategy."""
    strategy = HorovodStrategy()
    strategy.init()
    yield strategy
    strategy.clean_up()



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

import pytest
import yaml

from itwinai.parser import ConfigParser
from itwinai.pipeline import Pipeline
from itwinai.tests import (
    FakeGetterExec,
    FakeSaverExec,
    FakeSplitterExec,
    FakeTrainerExec,
)


def test_slice_into_sub_pipelines():
    """Test slicing the pipeline to obtain sub-pipelines as Pipeline objects."""
    p = Pipeline(["step1", "step2", "step3"])
    sub_pipe1, sub_pipe2 = p[:1], p[1:]
    assert isinstance(sub_pipe1, Pipeline)
    assert isinstance(sub_pipe2, Pipeline)
    assert len(sub_pipe1) == 1
    assert sub_pipe1[0] == "step1"
    assert len(sub_pipe2) == 2

    p = Pipeline(dict(step1="step1", step2="step2", step3="step3"))
    sub_pipe1, sub_pipe2 = p[:1], p[1:]
    assert isinstance(sub_pipe1, Pipeline)
    assert isinstance(sub_pipe2, Pipeline)
    assert len(sub_pipe1) == 1
    assert sub_pipe1["step1"] == "step1"
    assert len(sub_pipe2) == 2


def test_serialization_pipe_list():
    """Test dict serialization of pipeline defined as list of BaseComponent objects."""
    config = yaml.safe_load(pytest.PIPE_LIST_YAML)
    parser = ConfigParser(config=config)
    pipe = parser.parse_pipeline(pipeline_nested_key="my-list-pipeline")

    dict_pipe = pipe.to_dict()
    del dict_pipe["init_args"]["name"]
    dict_pipe = {"my-list-pipeline": dict_pipe}
    assert dict_pipe == config


def test_serialization_pipe_dict():
    """Test dict serialization of pipeline defined as dict of BaseComponent objects."""
    config = yaml.safe_load(pytest.PIPE_DICT_YAML)
    parser = ConfigParser(config=config)
    pipe = parser.parse_pipeline(pipeline_nested_key="my-dict-pipeline")

    dict_pipe = pipe.to_dict()
    del dict_pipe["init_args"]["name"]
    dict_pipe = {"my-dict-pipeline": dict_pipe}
    assert dict_pipe == config


def test_arguments_mismatch():
    """Test mismatch of arguments passed among components in a pipeline."""
    pipeline = Pipeline(
        [
            FakeGetterExec(data_uri="http://..."),
            FakeSplitterExec(train_prop=0.7),
            FakeTrainerExec(lr=1e-3, batch_size=32),
            # Adapter(policy=[f"{Adapter.INPUT_PREFIX}-1"]),
            FakeSaverExec(save_path="my_model.pth"),
        ]
    )
    with pytest.raises(TypeError) as exc_info:
        _ = pipeline.execute()
    assert "received too many input arguments" in str(exc_info.value)

    pipeline = Pipeline(
        [
            FakeGetterExec(data_uri="http://..."),
            FakeTrainerExec(lr=1e-3, batch_size=32),
        ]
    )
    with pytest.raises(TypeError) as exc_info:
        _ = pipeline.execute()
    assert "received too few input arguments" in str(exc_info.value)



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

import pytest

from itwinai.components import Adapter, Trainer
from itwinai.pipeline import Pipeline
from itwinai.serialization import SerializationError
from itwinai.tests import (
    FakeGetterExec,
    FakeSaverExec,
    FakeSplitterExec,
    FakeTrainerExec,
)


def test_serializable():
    """Test serialization of components."""
    comp = FakeGetterExec(data_uri="http://...")
    dict_serializ = comp.to_dict()
    assert isinstance(dict_serializ, dict)
    assert comp.name == "FakeGetterExec"
    assert dict_serializ == dict(
        class_path="itwinai.tests.dummy_components.FakeGetterExec",
        init_args=dict(data_uri="http://...", name=None),
    )

    # List
    comp = FakeGetterExec(data_uri=[1, 2, 3])
    dict_serializ = comp.to_dict()
    assert isinstance(dict_serializ, dict)
    assert comp.name == "FakeGetterExec"
    assert dict_serializ == dict(
        class_path="itwinai.tests.dummy_components.FakeGetterExec",
        init_args=dict(data_uri=[1, 2, 3], name=None),
    )

    # Tuple
    comp = FakeGetterExec(data_uri=(1, 2, 3))
    dict_serializ = comp.to_dict()
    assert isinstance(dict_serializ, dict)
    assert comp.name == "FakeGetterExec"
    assert dict_serializ == dict(
        class_path="itwinai.tests.dummy_components.FakeGetterExec",
        init_args=dict(data_uri=[1, 2, 3], name=None),
    )

    # Set
    comp = FakeGetterExec(data_uri={1, 2, 3})
    dict_serializ = comp.to_dict()
    assert isinstance(dict_serializ, dict)
    assert comp.name == "FakeGetterExec"
    assert dict_serializ == dict(
        class_path="itwinai.tests.dummy_components.FakeGetterExec",
        init_args=dict(data_uri=[1, 2, 3], name=None),
    )

    # Dict
    comp = FakeGetterExec(data_uri=dict(foo=12, bar="123", hl=3.14))
    dict_serializ = comp.to_dict()
    assert isinstance(dict_serializ, dict)
    assert comp.name == "FakeGetterExec"
    assert dict_serializ == dict(
        class_path="itwinai.tests.dummy_components.FakeGetterExec",
        init_args=dict(data_uri=dict(foo=12, bar="123", hl=3.14), name=None),
    )

    class NonSerializable:
        """Non-serializable object"""

    comp = FakeGetterExec(data_uri=NonSerializable())
    with pytest.raises(SerializationError) as exc_info:
        dict_serializ = comp.to_dict()
    assert "is not a Python built-in type and it does not extend" in str(exc_info.value)

    # Local component class
    class MyTrainer(Trainer):
        def execute(self): ...

    comp = MyTrainer()
    with pytest.raises(SerializationError) as exc_info:
        dict_serializ = comp.to_dict()
    assert "is defined locally, which is not supported for serialization." in str(
        exc_info.value
    )


def test_adapter():
    """Test Adapter component."""
    prefix = Adapter.INPUT_PREFIX
    adapter = Adapter(policy=[f"{prefix}{3-i}" for i in range(4)])
    result = adapter.execute(0, 1, 2, 3)
    assert result == (3, 2, 1, 0)

    result = adapter.execute(*tuple(range(100)))
    assert result == (3, 2, 1, 0)

    adapter = Adapter(policy=[f"{prefix}0" for i in range(4)])
    result = adapter.execute(0, 1, 2, 3)
    assert result == (0, 0, 0, 0)

    adapter = Adapter(policy=[f"{prefix}{i % 2}" for i in range(4)])
    result = adapter.execute(0, 1, 2, 3)
    assert result == (0, 1, 0, 1)

    adapter = Adapter(policy=[f"{prefix}2", "hello", "world", 3.14])
    result = adapter.execute(0, 1, 2, 3)
    assert result == (2, "hello", "world", 3.14)

    adapter = Adapter(policy=[1, 3, 5, 7, 11])
    result = adapter.execute(0, 1, 2, 3)
    assert result == (1, 3, 5, 7, 11)

    adapter = Adapter(policy=[f"{prefix}{9-i}" for i in range(10)])
    with pytest.raises(IndexError) as exc_info:
        result = adapter.execute(0, 1)
    assert str(exc_info.value) == (
        "The args received as input by 'Adapter' are not consistent with "
        "the given adapter policy because input args are too few! Input "
        "args are 2 but the policy foresees at least 10 items."
    )

    adapter = Adapter(policy=[])
    result = adapter.execute(*tuple(range(100)))
    assert result == ()


@pytest.mark.integration
def test_adapter_integration_pipeline():
    """Test integration of Adapter component in the pipeline,
    connecting other components.
    """
    pipeline = Pipeline(
        [
            FakeGetterExec(data_uri="http://..."),
            FakeSplitterExec(train_prop=0.7),
            FakeTrainerExec(lr=1e-3, batch_size=32),
            Adapter(policy=[f"{Adapter.INPUT_PREFIX}-1"]),
            FakeSaverExec(save_path="my_model.pth"),
        ]
    )
    saved_model = pipeline.execute()
    assert saved_model == FakeTrainerExec.model



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Jarl Sondre Sæther
#
# Credit:
# - Jarl Sondre Sæther <jarl.sondre.saether@cern.ch> - CERN
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

from typing import Tuple

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from scipy.constants import hour as SECONDS_IN_HOUR

matplotlib.use("Agg")


def calculate_average_gpu_utilization(gpu_data_df: pd.DataFrame) -> pd.DataFrame:
    """Calculates the average GPU utilization for each strategy and
    number of GPUs.

    Returns:
        pd.DataFrame: A DataFrame containing the average gpu utilization for
            each strategy and number of GPUs, with the columns ``strategy``,
            ``num_global_gpus`` and ``utilization``.
    """
    required_columns = {"strategy", "utilization", "num_global_gpus"}
    if not required_columns.issubset(set(gpu_data_df.columns)):
        missing_columns = set(required_columns) - set(gpu_data_df.columns)
        raise ValueError(
            f"DataFrame is missing the following columns: {missing_columns}"
        )

    utilization_data = []
    grouped_df = gpu_data_df.groupby(["strategy", "num_global_gpus"])
    for (strategy, num_gpus), group in grouped_df:
        utilization_data.append(
            {
                "strategy": strategy,
                "num_global_gpus": num_gpus,
                "utilization": group["utilization"].mean(),
            }
        )
    return pd.DataFrame(utilization_data)


def calculate_total_energy_expenditure(gpu_data_df: pd.DataFrame) -> pd.DataFrame:
    """Calculates the total energy expenditure in Watt hours for each strategy and
    number of GPUs. Expects that the existence of the appropriate DataFrame columns is
    handled before calling this function.

    Returns:
        pd.DataFrame: A DataFrame containing the total expenditure in Watt hours for
            each strategy and number of GPUs, with the columns ``strategy``,
            ``num_global_gpus`` and ``total_energy_wh``.
    """
    required_columns = {"strategy", "power", "num_global_gpus", "probing_interval"}
    if not required_columns.issubset(set(gpu_data_df.columns)):
        missing_columns = set(required_columns) - set(gpu_data_df.columns)
        raise ValueError(
            f"DataFrame is missing the following columns: {missing_columns}"
        )
    energy_data = []
    grouped_df = gpu_data_df.groupby(["strategy", "num_global_gpus"])
    for (strategy, num_gpus), group in grouped_df:

        if len(group["probing_interval"].unique()) != 1:
            raise ValueError(
                f"probing_interval must have the same value for each strategy and "
                f"number of GPUs, but was heterogeneous for strategy: {strategy} "
                f"and number of GPUs: {num_gpus}."
            )

        probing_interval = group["probing_interval"].iloc[0]
        total_energy_wh = group["power"].sum() * probing_interval / SECONDS_IN_HOUR
        energy_data.append(
            {
                "strategy": strategy,
                "num_global_gpus": num_gpus,
                "total_energy_wh": total_energy_wh,
            }
        )
    return pd.DataFrame(energy_data)


def gpu_bar_plot(
    data_df: pd.DataFrame, plot_title: str, y_label: str, main_column: str
) -> Tuple[Figure, Axes]:
    """Makes a bar plot of the given data for each strategy and number of GPUs.

    Args:
        data_df: The dataframe to extract the data from.
        plot_title: The title to give the plot.
        y_label: The label for the y-axis.
        main_column: The column to use for the height of the bar plot.
    """
    required_columns = {"strategy", "num_global_gpus", main_column}
    if not required_columns.issubset(set(data_df.columns)):
        missing_columns = set(required_columns) - set(data_df.columns)
        raise ValueError(
            f"DataFrame is missing the following columns: {missing_columns}"
        )

    sns.set_theme()

    strategies = data_df["strategy"].unique()
    unique_gpu_counts = np.array(data_df["num_global_gpus"].unique())

    fig, ax = plt.subplots()
    x = np.arange(len(unique_gpu_counts))

    bar_width = 1 / (len(strategies) + 1)
    static_offset = (len(strategies) - 1) / 2
    for strategy_idx, strategy in enumerate(strategies):
        dynamic_bar_offset = strategy_idx - static_offset
        strategy_data = data_df[data_df["strategy"] == strategy]

        # Ensuring the correct spacing of the bars
        strategy_num_gpus = len(strategy_data["num_global_gpus"])

        ax.bar(
            x=x[:strategy_num_gpus] + dynamic_bar_offset * bar_width,
            height=strategy_data[main_column],
            width=bar_width,
            label=strategy,
        )

    ax.set_xlabel("Num GPUs")
    ax.set_ylabel(y_label)
    ax.set_title(plot_title)
    ax.set_xticks(x)
    ax.set_xticklabels(unique_gpu_counts)
    ax.legend(title="Strategy")

    figure_width = max(int(2 * len(unique_gpu_counts)), 8)
    fig.set_figwidth(figure_width)
    fig.set_figheight(figure_width * 0.8)

    sns.reset_orig()

    return fig, ax



#!/usr/bin/env python
# Derived from (restricted access):
# https://gitlab.cern.ch/escape-wp2/docker-images/-/blob/master/datalake-singleuser/bin/configure.py

import configparser
import json
import os

# HOME = '/home/jovyan'
# HOME = '/ceph/hpc/home/ciangottinid'


def write_jupyterlab_config():
    HOME = os.getenv('HOME', '/ceph/hpc/home/ciangottinid')

    file_path = HOME + '/.jupyter/jupyter_notebook_config.json'
    if not os.path.isfile(file_path):
        os.makedirs(HOME + '/.jupyter/', exist_ok=True)
    else:
        config_file = open(file_path, 'r')
        config_payload = config_file.read()
        config_file.close()

    try:
        config_json = json.loads(config_payload)
    except Exception:
        config_json = {}

# Looking to the rucio-jupyterlab configuration;
# https://github.com/rucio/jupyterlab-extension/blob/master/rucio_jupyterlab/config/schema.py#L101
#  either ("destination_rse", "rse_mount_path") either ("rucio_ca_cert") are required env
# vars, even if they are defined in the jhub manifest.
# Adding 'rucio_base_url' too - from debugging experience

    # instance_config = {
    #     "name": os.getenv('RUCIO_NAME', 'default'),
    #     "display_name": os.getenv('RUCIO_DISPLAY_NAME', 'Default Instance'),
    #     "rucio_base_url": os.getenv('RUCIO_BASE_URL', 'DEFAULT rucio base url'),
    #     "rucio_auth_url": os.getenv('RUCIO_AUTH_URL'),
    #     "rucio_webui_url": os.getenv('RUCIO_WEBUI_URL'),
    #     "rucio_ca_cert": os.getenv('RUCIO_CA_CERT', 'cacert.pem'),
    #     "site_name": os.getenv('RUCIO_SITE_NAME'),
    #     "vo": os.getenv('RUCIO_VO'),
    #     "voms_enabled": os.getenv('RUCIO_VOMS_ENABLED', '0') == '1',
    #     "voms_vomses_path": os.getenv('RUCIO_VOMS_VOMSES_PATH'),
    #     "voms_certdir_path": os.getenv('RUCIO_VOMS_CERTDIR_PATH'),
    #     "voms_vomsdir_path": os.getenv('RUCIO_VOMS_VOMSDIR_PATH'),
    #     "destination_rse": os.getenv('RUCIO_DESTINATION_RSE', 'DEFAULT rse destination'),
    #     "rse_mount_path": os.getenv('RUCIO_RSE_MOUNT_PATH', 'DEFAULT rse mount path'),
    #     "replication_rule_lifetime_days": int(os.getenv(
        # 'RUCIO_REPLICATION_RULE_LIFETIME_DAYS')) if os.getenv(
        # 'RUCIO_REPLICATION_RULE_LIFETIME_DAYS') else None,
    #     "path_begins_at": int(os.getenv('RUCIO_PATH_BEGINS_AT', '0')),
    #     "mode": os.getenv('RUCIO_MODE', 'replica'),
    #     "wildcard_enabled": os.getenv('RUCIO_WILDCARD_ENABLED', '0') == '1',
    #     "oidc_auth": os.getenv('RUCIO_OIDC_AUTH'),
    #     "oidc_env_name": os.getenv('RUCIO_OIDC_ENV_NAME'),
    #     "oidc_file_name": os.getenv('RUCIO_OIDC_FILE_NAME'),
    # }

    # instance_config = {k: v for k,
    #                    v in instance_config.items() if v is not None}
    # config_json['RucioConfig'] = {
    #     'instances': [instance_config],
    #     "default_instance": os.getenv('RUCIO_DEFAULT_INSTANCE',),
    #     "default_auth_type": os.getenv('RUCIO_DEFAULT_AUTH_TYPE'),
    # }

    # Debugging and hardcoding from here

    instance_config = {
        "name": "rucio-intertwin-testbed.desy.de",
        "display_name": "interTwinRUCIO",
        "rucio_base_url": "https://rucio-intertwin-testbed.desy.de",
        "rucio_auth_url": "https://rucio-intertwin-testbed-auth.desy.de",
        "rucio_ca_cert": "/opt/conda/lib/python3.9/site-packages/certifi/cacert.pem",
        "site_name": "VEGA",
        "voms_enabled": os.getenv('RUCIO_VOMS_ENABLED', '0') == '1',
        "destination_rse": "VEGA-DCACHE",
        "rse_mount_path": "/dcache/sling.si/projects/intertwin",
        "path_begins_at": 4,
        "mode": "replica",
        # "mode": "download",
        "wildcard_enabled": os.getenv('RUCIO_WILDCARD_ENABLED', '0') == '0',
        "oidc_auth": "env",
        "oidc_env_name": "RUCIO_ACCESS_TOKEN"
    }

    instance_config = {k: v for k,
                       v in instance_config.items() if v is not None}
    config_json['RucioConfig'] = {
        'instances': [instance_config],
        "default_instance": os.getenv('RUCIO_DEFAULT_INSTANCE',
                                      'rucio-intertwin-testbed.desy.de'),
        "default_auth_type": os.getenv('RUCIO_DEFAULT_AUTH_TYPE', 'oidc'),
    }

    # up to here

    config_file = open(file_path, 'w')
    config_file.write(json.dumps(config_json, indent=2))
    config_file.close()


def write_rucio_config():

    rucio_config = configparser.ConfigParser()

    client_config = {
        'rucio_host': os.getenv('RUCIO_BASE_URL',
                                'https://rucio-intertwin-testbed.desy.de'),
        'auth_host': os.getenv('RUCIO_AUTH_URL',
                               'https://rucio-intertwin-testbed-auth.desy.de'),
        'ca_cert': os.getenv('RUCIO_CA_CERT', '/certs/rucio_ca.pem'),
        'auth_type': os.getenv('RUCIO_AUTH_TYPE', 'oidc'),  # 'x509' or 'oidc'
        # This is the RUCIO account name, need to be mapped from idp
        'account': os.getenv('RUCIO_ACCOUNT', '$RUCIO_ACCOUNT'),
        'oidc_polling': 'true',
        'oidc_scope': 'openid profile offline_access eduperson_entitlement',
        # 'username': os.getenv('RUCIO_USERNAME', ''),
        # 'password': os.getenv('RUCIO_PASSWORD', ''),
        'auth_token_file_path': '/tmp/rucio_oauth.token',
        'request_retries': 3,
        'protocol_stat_retries': 6
    }
    client_config = dict((k, v) for k, v in client_config.items() if v)
    rucio_config['client'] = client_config

    if not os.path.isfile('/opt/rucio/etc/rucio.cfg'):
        os.makedirs('/opt/rucio/etc/', exist_ok=True)

    with open('/opt/rucio/etc/rucio.cfg', 'w') as f:
        rucio_config.write(f)


if __name__ == '__main__':
    write_jupyterlab_config()
    # write_rucio_config()



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Anna Lappe
#
# Credit:
# - Anna Lappe <anna.elisa.lappe@cern.ch> - CERN
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

from typing import Dict, Literal

import numpy as np
import torch
from ray import train
from torch.nn import CrossEntropyLoss
from torch.optim import Adam
from torchvision.models import resnet18

from itwinai.loggers import Logger
from itwinai.torch.config import TrainingConfiguration
from itwinai.torch.distributed import DeepSpeedStrategy
from itwinai.torch.trainer import TorchTrainer


class MyTrainer(TorchTrainer):
    def __init__(
        self,
        epochs: int,
        config: Dict | TrainingConfiguration,
        strategy: Literal["ddp", "deepspeed", "horovod"] = "ddp",
        name: str | None = None,
        logger: Logger | None = None,
    ) -> None:
        self.config = config
        super().__init__(
            config=config, epochs=epochs, strategy=strategy, name=name, logger=logger
        )

    def create_model_loss_optimizer(self):
        model = resnet18(num_classes=10)
        model.conv1 = torch.nn.Conv2d(
            1, 64, kernel_size=(7, 7), stride=(2, 2), padding=(3, 3), bias=False
        )
        # First, define strategy-wise optional configurations
        if isinstance(self.strategy, DeepSpeedStrategy):
            distribute_kwargs = dict(
                config_params=dict(train_micro_batch_size_per_gpu=self.config.batch_size)
            )
        else:
            distribute_kwargs = {}
        optimizer = Adam(model.parameters(), lr=self.config.optim_lr)
        self.model, self.optimizer, _ = self.strategy.distributed(
            model, optimizer, **distribute_kwargs
        )
        self.loss = CrossEntropyLoss()

    def train(self):
        for epoch in range(self.epochs):
            if self.strategy.global_world_size() > 1:
                self.set_epoch(epoch)

            train_losses = []
            val_losses = []

            for images, labels in self.train_dataloader:
                device = self.strategy.device()
                images, labels = images.to(device), labels.to(device)

                outputs = self.model(images)
                train_loss = self.loss(outputs, labels)
                self.optimizer.zero_grad()
                train_loss.backward()
                self.optimizer.step()
                train_losses.append(train_loss.detach().cpu().numpy())

            for images, labels in self.validation_dataloader:
                device = self.strategy.device()
                images, labels = images.to(device), labels.to(device)

                with torch.no_grad():
                    outputs = self.model(images)
                    val_loss = self.loss(outputs, labels)
                val_losses.append(val_loss.detach().cpu().numpy())

            epoch_val_loss = np.mean(val_losses)
            epoch_train_loss = np.mean(train_losses)

            self.log(epoch_train_loss, "train_loss", kind="metric", step=epoch)
            self.log(epoch_val_loss, "val_loss", kind="metric", step=epoch)

            # Report training metrics of last epoch to Ray
            train.report({"loss": epoch_val_loss})



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

"""Default configuration"""

from typing import Literal

from pydantic import BaseModel


class Configuration(BaseModel, extra="allow"):
    """Base configuration class."""

    def __getitem__(self, idx):
        return self.__getattribute__(idx)


class TrainingConfiguration(Configuration):
    """Default configuration object for training.
    Override and/or create new configurations using the constructor.

    Example:

    >>> cfg = TrainingConfiguration(batch_size=17, param_a=42)
    >>> print(cfg.batch_size)  # returns 17 (overrides default)
    >>> print(cfg.param_a)     # returns 42 (new value)
    >>> print(cfg.pin_memory)  # returns the default value
    >>>
    >>> from rich import print
    >>> print(cfg)             # pretty-print of configuration

    .. warning::

        Don't reinvent parameters that already exist in the training coniguration, if possible.
        Instead, use the name of the parameters in the training configuration when possible to
        avoid inconsistencies. For instance, the training configuration defines the learning
        rate as ``optim_lr``, so if you redefine it as ``lr`` by doing
        ``TrainingConfiguration(lr=0.005)`` in the configuration you will now have both
        ``optim_lr`` (created by default) and ``lr`` (created by you). This may create
        confusion and potentially (and silently) break the logic in your code.
    """

    #: Batch size. In a distributed environment it is usually the
    #: per-worker batch size. Defaults to 32.
    batch_size: int = 32
    #: Whether to shuffle train dataset when creating a torch ``DataLoader``.
    #: Defaults to False.
    shuffle_train: bool = False
    #: Whether to shuffle validation dataset when creating a torch
    #: ``DataLoader``.
    #: Defaults to False.
    shuffle_validation: bool = False
    #: Whether to shuffle test dataset when creating a torch ``DataLoader``.
    #: Defaults to False.
    shuffle_test: bool = False
    #: Whether to pin GPU memory. Property of torch ``DataLoader``.
    #: Defaults to False.
    pin_gpu_memory: bool = False
    #: Number of parallel workers used by torch ``DataLoader``.
    #: Defaults to 4.
    num_workers_dataloader: int = 4
    #: Loss function. Defaults to 'cross_entropy'
    loss: Literal["mse", "nllloss", "cross_entropy", "l1", "l2"] = "cross_entropy"
    #: Name of the optimizer to use. Defaults to 'adam'.
    optimizer: Literal["adadelta", "adam", "rmsprop", "sgd"] = "adam"
    #: Learning rate used by the optimizer. Defaults to 1e-3.
    optim_lr: float = 1e-3
    #: Momentum used by some optimizers (e.g., SGD). Defaults to 0.9.
    optim_momentum: float = 0.9
    #: Weight decay parameter for the optimizer. Defaults to 0.
    optim_weight_decay: float = 0.0
    #: Parameter of Horovod's ``DistributedOptimizer``: uses float16
    #: operations in the allreduce
    #: distributed gradients aggregation. Better performances at
    #: lower precision. Defaults to False.
    fp16_allreduce: bool = False
    #: Parameter of Horovod's ``DistributedOptimizer``: use Adasum
    #: optimization.
    #: Defaults to False.
    use_adasum: bool = False
    #: Parameter of Horovod's ``DistributedOptimizer``: scale
    #: gradients before adding them up.
    #: Defaults to 1.0.
    gradient_predivide_factor: float = 1.0
    #: Torch distributed
    #: `backend <https://pytorch.org/docs/stable/distributed.html#backends>`_.
    #: Defaults to ``nccl``.
    dist_backend: Literal["nccl", "gloo", "mpi"] = "nccl"



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# - Anna Lappe <anna.elisa.lappe@cern.ch> - CERN
# - Jarl Sondre Sæther <jarl.sondre.saether@cern.ch> - CERN
# - Henry Mutegeki <henry.mutegeki@cern.ch> - CERN
# --------------------------------------------------------------------------------------


"""Provides training logic for PyTorch models via Trainer classes."""

import os
import sys
import tempfile
from collections import defaultdict
from pathlib import Path
from typing import Any, Callable, Dict, Iterable, List, Literal, Optional, Tuple, Union

import torch
import torch.distributed as dist
import torch.nn as nn
import torch.optim as optim
import torchvision
from torch.nn.parallel import DistributedDataParallel as DDP
from torch.optim.optimizer import Optimizer
from torch.utils.data import DataLoader, Dataset, Sampler
from torch.utils.data.distributed import DistributedSampler
from tqdm import tqdm

from ..components import Trainer, monitor_exec
from ..loggers import Logger, LogMixin
from ..utils import load_yaml
from .config import TrainingConfiguration
from .distributed import (
    DeepSpeedStrategy,
    HorovodStrategy,
    NonDistributedStrategy,
    RayDDPStrategy,
    RayDeepSpeedStrategy,
    TorchDDPStrategy,
    TorchDistributedStrategy,
    distributed_resources_available,
)
from .reproducibility import seed_worker, set_seed
from .type import Batch, LrScheduler, Metric


class TorchTrainer(Trainer, LogMixin):
    """Trainer class for torch training algorithms.

    Args:
        config (Union[Dict, TrainingConfiguration]): training configuration
            containing hyperparameters.
        epochs (int): number of training epochs.
        model (Optional[Union[nn.Module, str]], optional): pytorch model to
            train or a string identifier. Defaults to None.
        strategy (Literal['ddp', 'deepspeed', 'horovod'], optional):
            distributed strategy. Defaults to 'ddp'.
        validation_every (Optional[int], optional): run a validation epoch
            every ``validation_every`` epochs. Disabled if None. Defaults to 1.
        test_every (Optional[int], optional): run a test epoch
            every ``test_every`` epochs. Disabled if None. Defaults to None.
        random_seed (Optional[int], optional): set random seed for
            reproducibility. If None, the seed is not set. Defaults to None.
        logger (Optional[Logger], optional): logger for ML tracking.
            Defaults to None.
        metrics (Optional[Dict[str, Metric]], optional): map of torchmetrics
            metrics. Defaults to None.
        checkpoints_location (str): path to checkpoints directory.
            Defaults to "checkpoints".
        checkpoint_every (Optional[int]): save a checkpoint every
            ``checkpoint_every`` epochs. Disabled if None. Defaults to None.
        disable_tqdm (bool): whether to disable tqdm progress bar(s).
        name (Optional[str], optional): trainer custom name. Defaults to None.
        profiling_wait_epochs (int): how many epochs to wait before starting
            the profiler.
        profiling_warmup_epochs (int): length of the profiler warmup phase in terms of
            number of epochs.
    """

    # TODO:
    #   - extract BaseTorchTrainer and extend it creating a set of trainer
    #     templates (e.g.. GAN, Classifier, Transformer) allowing scientists
    #     to reuse ML algos.

    _strategy: TorchDistributedStrategy = None

    #: PyTorch ``DataLoader`` for training dataset.
    train_dataloader: DataLoader = None
    #: PyTorch ``DataLoader`` for validation dataset.
    validation_dataloader: DataLoader = None
    #: PyTorch ``DataLoader`` for test dataset.
    test_dataloader: DataLoader = None
    #: PyTorch model to train.
    model: nn.Module = None
    #: Loss criterion.
    loss: Callable = None
    #: Optimizer.
    optimizer: Optimizer = None
    #: Learning rate scheduler.
    lr_scheduler: LrScheduler = None
    #: PyTorch random number generator (PRNG).
    torch_rng: torch.Generator = None
    #: itwinai ``itwinai.Logger``
    logger: Logger = None
    #: Total number training batches used so far, across all epochs.
    train_glob_step: int = 0
    #: Total number validation batches used so far, across all epochs.
    validation_glob_step: int = 0
    #: Total number test batches used so far, across all epochs.
    test_glob_step: int = 0
    #: Dictionary of ``torchmetrics`` metrics, indexed by user-defined names.
    metrics: Dict[str, Metric]
    #: PyTorch Profiler for communication vs. computation comparison
    profiler: Optional[Any]

    def __init__(
        self,
        config: Union[Dict, TrainingConfiguration],
        epochs: int,
        model: Optional[Union[nn.Module, str]] = None,
        strategy: Optional[Literal["ddp", "deepspeed", "horovod"]] = "ddp",
        validation_every: Optional[int] = 1,
        test_every: Optional[int] = None,
        random_seed: Optional[int] = None,
        logger: Optional[Logger] = None,
        metrics: Optional[Dict[str, Metric]] = None,
        checkpoints_location: str = "checkpoints",
        checkpoint_every: Optional[int] = None,
        disable_tqdm: bool = False,
        name: Optional[str] = None,
        profiling_wait_epochs: int = 1,
        profiling_warmup_epochs: int = 2,
    ) -> None:
        super().__init__(name)
        self.save_parameters(**self.locals2params(locals()))

        # config is mean to store all hyperparameters, which can very from use
        # case to use case and include learning_rate, batch_size....
        if isinstance(config, dict):
            config = TrainingConfiguration(**config)

        self.config = config
        self.epochs = epochs
        self.model = model
        self.strategy = strategy
        self.validation_every = validation_every
        self.test_every = test_every
        self.random_seed = random_seed
        self.logger = logger
        self.metrics = metrics if metrics is not None else {}
        self.checkpoints_location = checkpoints_location
        os.makedirs(self.checkpoints_location, exist_ok=True)
        self.checkpoint_every = checkpoint_every
        self.disable_tqdm = disable_tqdm
        self.profiler = None
        self.profiling_wait_epochs = profiling_wait_epochs
        self.profiling_warmup_epochs = profiling_warmup_epochs

    @property
    def strategy(self) -> TorchDistributedStrategy:
        """Strategy currently in use."""
        return self._strategy

    @strategy.setter
    def strategy(self, strategy: str | TorchDistributedStrategy) -> None:
        if isinstance(strategy, TorchDistributedStrategy):
            self._strategy = strategy
        else:
            self._strategy = self._detect_strategy(strategy)

    @property
    def device(self) -> str:
        """Current device from distributed strategy."""
        return self.strategy.device()

    def _detect_strategy(self, strategy: str) -> TorchDistributedStrategy:
        if strategy is None or not distributed_resources_available():
            print("WARNING: falling back to non-distributed strategy.")
            strategy_obj = NonDistributedStrategy()
        elif strategy == "ddp":
            strategy_obj = TorchDDPStrategy(backend=self.config.dist_backend)
        elif strategy == "horovod":
            strategy_obj = HorovodStrategy()
        elif strategy == "deepspeed":
            strategy_obj = DeepSpeedStrategy(backend=self.config.dist_backend)
        else:
            raise NotImplementedError(f"Strategy '{strategy}' is not recognized/implemented.")
        return strategy_obj

    def _init_distributed_strategy(self) -> None:
        if not self.strategy.is_initialized:
            self.strategy.init()

    def _optimizer_from_config(self) -> None:
        if self.config.optimizer == "adadelta":
            self.optimizer = optim.Adadelta(
                self.model.parameters(),
                lr=self.config.optim_lr,
                weight_decay=self.config.optim_weight_decay,
            )
        elif self.config.optimizer == "adam":
            self.optimizer = optim.Adam(
                self.model.parameters(),
                lr=self.config.optim_lr,
                weight_decay=self.config.optim_weight_decay,
            )
        elif self.config.optimizer == "rmsprop":
            self.optimizer = optim.RMSprop(
                self.model.parameters(),
                lr=self.config.optim_lr,
                weight_decay=self.config.optim_weight_decay,
                momentum=self.config.optim_momentum,
            )
        elif self.config.optimizer == "sgd":
            self.optimizer = optim.SGD(
                self.model.parameters(),
                lr=self.config.optim_lr,
                weight_decay=self.config.optim_weight_decay,
                momentum=self.config.optim_momentum,
            )
        else:
            raise ValueError(
                "Unrecognized self.config.optimizer! Check the docs for "
                "supported values and consider overriding "
                "create_model_loss_optimizer method for more flexibility."
            )

    def _loss_from_config(self) -> None:
        if self.config.loss == "nllloss":
            self.loss = nn.functional.nll_loss
        elif self.config.loss == "cross_entropy":
            self.loss = nn.functional.cross_entropy
        elif self.config.loss == "mse":
            self.loss = nn.functional.mse_loss
        else:
            raise ValueError(
                "Unrecognized self.config.loss! Check the docs for "
                "supported values and consider overriding "
                "create_model_loss_optimizer method for more flexibility."
            )

    def get_default_distributed_kwargs(self) -> Dict:
        """Gives the default kwargs for the trainer's strategy's distributed() method."""

        if isinstance(self.strategy, DeepSpeedStrategy):
            # Batch size definition is not optional for DeepSpeedStrategy!
            distribute_kwargs = dict(
                config_params=dict(train_micro_batch_size_per_gpu=self.config.batch_size)
            )
        elif isinstance(self.strategy, HorovodStrategy):
            import horovod.torch as hvd

            distribute_kwargs = dict(
                compression=(
                    hvd.Compression.fp16
                    if self.config.fp16_allreduce
                    else hvd.Compression.none
                ),
                op=hvd.Adasum if self.config.use_adasum else hvd.Average,
                gradient_predivide_factor=self.config.gradient_predivide_factor,
            )
        else:
            distribute_kwargs = {}

        return distribute_kwargs

    def create_model_loss_optimizer(self) -> None:
        """
        Instantiate a torch model, loss, optimizer, and LR scheduler using the
        configuration provided in the Trainer constructor.
        Generally a user-defined method.
        """
        ###################################
        # Dear user, this is a method you #
        # may be interested to override!  #
        ###################################

        if self.model is None:
            raise ValueError(
                "self.model is None! Either pass it to the constructor or "
                "override create_model_loss_optimizer method."
            )

        # Parse optimizer from training configuration
        # Optimizer can be changed with a custom one here!
        self._optimizer_from_config()

        # Parse loss from training configuration
        # Loss can be changed with a custom one here!
        self._loss_from_config()

        # IMPORTANT: model, optimizer, and scheduler need to be distributed
        distribute_kwargs = self.get_default_distributed_kwargs()

        # Distributed model, optimizer, and scheduler
        (self.model, self.optimizer, self.lr_scheduler) = self.strategy.distributed(
            self.model, self.optimizer, self.lr_scheduler, **distribute_kwargs
        )

    def create_dataloaders(
        self,
        train_dataset: Dataset,
        validation_dataset: Optional[Dataset] = None,
        test_dataset: Optional[Dataset] = None,
    ) -> None:
        """
        Create train, validation and test dataloaders using the
        configuration provided in the Trainer constructor.
        Generally a user-defined method.

        Args:
            train_dataset (Dataset): training dataset object.
            validation_dataset (Optional[Dataset]): validation dataset object.
                Default None.
            test_dataset (Optional[Dataset]): test dataset object.
                Default None.
        """

        ###################################
        # Dear user, this is a method you #
        # may be interested to override!  #
        ###################################
        self.train_dataloader = self.strategy.create_dataloader(
            dataset=train_dataset,
            batch_size=self.config.batch_size,
            num_workers=self.config.num_workers_dataloader,
            pin_memory=self.config.pin_gpu_memory,
            generator=self.torch_rng,
            shuffle=self.config.shuffle_train,
        )
        if validation_dataset is not None:
            self.validation_dataloader = self.strategy.create_dataloader(
                dataset=validation_dataset,
                batch_size=self.config.batch_size,
                num_workers=self.config.num_workers_dataloader,
                pin_memory=self.config.pin_gpu_memory,
                generator=self.torch_rng,
                shuffle=self.config.shuffle_validation,
            )
        if test_dataset is not None:
            self.test_dataloader = self.strategy.create_dataloader(
                dataset=test_dataset,
                batch_size=self.config.batch_size,
                num_workers=self.config.num_workers_dataloader,
                pin_memory=self.config.pin_gpu_memory,
                generator=self.torch_rng,
                shuffle=self.config.shuffle_test,
            )

    def _setup_metrics(self):
        """Move metrics to current device."""
        for m_name, metric in self.metrics.items():
            self.metrics[m_name] = metric.to(self.device)

    @monitor_exec
    def execute(
        self,
        train_dataset: Dataset,
        validation_dataset: Optional[Dataset] = None,
        test_dataset: Optional[Dataset] = None,
    ) -> Tuple[Dataset, Dataset, Dataset, Any]:
        """Prepares distributed environment and data structures
        for the actual training.

        Args:
            train_dataset (Dataset): training dataset.
            validation_dataset (Optional[Dataset], optional): validation
                dataset. Defaults to None.
            test_dataset (Optional[Dataset], optional): test dataset.
                Defaults to None.

        Returns:
            Tuple[Dataset, Dataset, Dataset, Any]: training dataset,
            validation dataset, test dataset, trained model.
        """
        self.torch_rng = set_seed(self.random_seed)
        self._init_distributed_strategy()
        self._setup_metrics()

        self.create_dataloaders(
            train_dataset=train_dataset,
            validation_dataset=validation_dataset,
            test_dataset=test_dataset,
        )
        self.create_model_loss_optimizer()

        if self.logger:
            self.logger.create_logger_context(rank=self.strategy.global_rank())
            hparams = self.config.model_dump()
            hparams["distributed_strategy"] = self.strategy.__class__.__name__
            self.logger.save_hyperparameters(hparams)

        self.train()

        if self.logger:
            self.logger.destroy_logger_context()
        self.strategy.clean_up()
        return train_dataset, validation_dataset, test_dataset, self.model

    def _set_epoch_dataloaders(self, epoch: int):
        """
        Sets epoch in the distributed sampler of a dataloader when using it.
        """
        if self.strategy.is_distributed:
            self.train_dataloader.sampler.set_epoch(epoch)
            if self.validation_dataloader is not None:
                self.validation_dataloader.sampler.set_epoch(epoch)
            if self.test_dataloader is not None:
                self.test_dataloader.sampler.set_epoch(epoch)

    def set_epoch(self, epoch: int) -> None:
        """Set current epoch at the beginning of training.

        Args:
            epoch (int): epoch number, from 0 to ``epochs-1``.
        """
        if self.profiler is not None and epoch > 0:
            # We don't want to start stepping until after the first epoch
            self.profiler.step()
        self._set_epoch_dataloaders(epoch)

    def log(
        self,
        item: Union[Any, List[Any]],
        identifier: Union[str, List[str]],
        kind: str = "metric",
        step: Optional[int] = None,
        batch_idx: Optional[int] = None,
        **kwargs,
    ) -> None:
        """Log ``item`` with ``identifier`` name of ``kind`` type at ``step``
        time step.

        Args:
            item (Union[Any, List[Any]]): element to be logged (e.g., metric).
            identifier (Union[str, List[str]]): unique identifier for the
                element to log(e.g., name of a metric).
            kind (str, optional): type of the item to be logged. Must be one
                among the list of self.supported_types. Defaults to 'metric'.
            step (Optional[int], optional): logging step. Defaults to None.
            batch_idx (Optional[int], optional): DataLoader batch counter
                (i.e., batch idx), if available. Defaults to None.
        """
        if self.logger:
            self.logger.log(
                item=item,
                identifier=identifier,
                kind=kind,
                step=step,
                batch_idx=batch_idx,
                **kwargs,
            )

    def save_checkpoint(
        self, name: str, epoch: int, loss: Optional[torch.Tensor] = None
    ) -> None:
        """Save training checkpoint.

        Args:
            name (str): name of the checkpoint.
            epoch (int): current training epoch.
            loss (Optional[torch.Tensor]): current loss (if available).
        """
        state = dict(
            epoch=epoch,
            loss=loss,
            optimizer=self.optimizer.state_dict(),
            model=self.model.state_dict(),
            lr_scheduler=self.lr_scheduler,
        )
        ckpt_path = os.path.join(self.checkpoints_location, name)
        torch.save(state, ckpt_path)
        # print(f"Saved '{name}' checkpoint at {ckpt_path}")

        # Save checkpoint to logger
        self.log(ckpt_path, name, kind="artifact")

    def load_checkpoint(self, name: str) -> None:
        """Load state from a checkpoint.

        Args:
            name (str): name of the checkpoint to load, assuming it
                is under ``self.checkpoints_location`` location.
        """
        ckpt_path = os.path.join(self.checkpoints_location, name)
        state = torch.load(ckpt_path, map_location=self.device)
        self.model.load_state_dict(state["model"])
        self.optimizer.load_state_dict(state["optimizer"])
        self.lr_scheduler = state["lr_scheduler"]

    def compute_metrics(
        self,
        true: Batch,
        pred: Batch,
        logger_step: int,
        batch_idx: Optional[int],
        stage: str = "train",
    ) -> Dict[str, Any]:
        """Compute and log metrics.

        Args:
            metrics (Dict[str, Metric]): metrics dict. Can be
                ``self.train_metrics`` or ``self.validation_metrics``.
            true (Batch): true values.
            pred (Batch): predicted values.
            logger_step (int): global step to pass to the logger.
            stage (str): 'train', 'validation'...

        Returns:
            Dict[str, Any]: metric values.
        """
        m_values = {}
        for m_name, metric in self.metrics.items():
            # metric = metric.to(self.device)
            m_val = metric(pred, true).detach().cpu().numpy()
            self.log(
                item=m_val,
                identifier=f"{stage}_{m_name}",
                kind="metric",
                step=logger_step,
                batch_idx=batch_idx,
            )
            m_values[m_name] = m_val
        return m_values

    def train(self):
        """Trains a machine learning model.
        Main training loop/logic.

        Args:
            train_dataset (Dataset): training dataset.
            validation_dataset (Dataset): validation dataset.
            test_dataset (Dataset): test dataset.

        Returns:
            Tuple[Dataset, Dataset, Dataset, Any]: training dataset,
            validation dataset, test dataset, trained model.
        """
        best_loss = float("inf")

        progress_bar = tqdm(
            range(self.epochs),
            desc="Epochs",
            disable=self.disable_tqdm or not self.strategy.is_main_worker,
        )

        for epoch in progress_bar:
            progress_bar.set_description(f"Epoch {epoch + 1}/{self.epochs}")

            epoch_n = epoch + 1
            self.set_epoch(epoch)
            self.train_epoch(epoch)
            if self.validation_every and epoch_n % self.validation_every == 0:
                val_loss = self.validation_epoch(epoch)

                # Checkpointing current best model
                worker_val_losses = self.strategy.gather(val_loss, dst_rank=0)
                if self.strategy.is_main_worker:
                    avg_loss = torch.mean(torch.stack(worker_val_losses)).detach().cpu()
                    if avg_loss < best_loss and self.checkpoint_every is not None:
                        ckpt_name = "best_model.pth"
                        self.save_checkpoint(name=ckpt_name, epoch=epoch, loss=avg_loss)
                        best_loss = avg_loss

            if self.test_every and epoch_n % self.test_every == 0:
                self.test_epoch(epoch)

            # Periodic checkpointing
            if (
                self.strategy.is_main_worker
                and self.checkpoint_every
                and epoch_n % self.checkpoint_every == 0
            ):
                ckpt_name = f"epoch_{epoch}.pth"
                self.save_checkpoint(name=ckpt_name, epoch=epoch)

    def train_epoch(self, epoch: int) -> torch.Tensor:
        """Perform a complete sweep over the training dataset, completing an
        epoch of training.

        Args:
            epoch (int): current epoch number, from 0 to ``self.epochs - 1``.

        Returns:
            Loss: average training loss for the current epoch.
        """
        self.model.train()
        train_loss_sum = 0.0
        train_metrics_sum = defaultdict(float)
        batch_counter = 0

        progress_bar = tqdm(
            enumerate(self.train_dataloader),
            total=len(self.train_dataloader) // self.strategy.global_world_size(),
            desc="Train batches",
            disable=self.disable_tqdm or not self.strategy.is_main_worker,
            leave=False,  # Set this to true to see how many batches were used
        )

        for batch_idx, train_batch in progress_bar:
            loss, metrics = self.train_step(batch=train_batch, batch_idx=batch_idx)
            train_loss_sum += loss
            batch_counter += 1
            for name, val in metrics.items():
                train_metrics_sum[name] += val

            # Important: update counter
            self.train_glob_step += 1

        # Aggregate and log losses
        avg_loss = train_loss_sum / batch_counter
        self.log(
            item=avg_loss.item(),
            identifier="train_loss_epoch",
            kind="metric",
            step=self.train_glob_step,
        )
        # Aggregate and log metrics
        for m_name, m_val in train_metrics_sum.items():
            self.log(
                item=m_val / batch_counter,
                identifier="train_" + m_name + "_epoch",
                kind="metric",
                step=self.train_glob_step,
            )

        return avg_loss

    def train_step(self, batch: Batch, batch_idx: int) -> Tuple[torch.Tensor, Dict[str, Any]]:
        """Perform a single optimization step using a batch sampled from the
        training dataset.

        Args:
            batch (Batch): batch sampled by a dataloader.
            batch_idx (int): batch index in the dataloader.

        Returns:
            Tuple[Loss, Dict[str, Any]]: batch loss and dictionary of metric
            values with the same structure of ``self.metrics``.
        """
        x, y = batch
        x, y = x.to(self.device), y.to(self.device)

        self.optimizer.zero_grad()
        pred_y = self.model(x)
        loss = self.loss(pred_y, y)
        loss.backward()
        self.optimizer.step()

        # Log metrics
        self.log(
            item=loss.item(),
            identifier="train_loss",
            kind="metric",
            step=self.train_glob_step,
            batch_idx=batch_idx,
        )
        metrics: Dict[str, Any] = self.compute_metrics(
            true=y,
            pred=pred_y,
            logger_step=self.train_glob_step,
            batch_idx=batch_idx,
            stage="train",
        )
        return loss, metrics

    def validation_epoch(self, epoch: int) -> torch.Tensor:
        """Perform a complete sweep over the validation dataset, completing an
        epoch of validation.

        Args:
            epoch (int): current epoch number, from 0 to ``self.epochs - 1``.

        Returns:
            Optional[Loss]: average validation loss for the current epoch if
                self.validation_dataloader is not None
        """
        if self.validation_dataloader is None:
            return

        progress_bar = tqdm(
            enumerate(self.validation_dataloader),
            total=len(self.validation_dataloader) // self.strategy.global_world_size(),
            desc="Validation batches",
            disable=self.disable_tqdm or not self.strategy.is_main_worker,
            leave=False,  # Set this to true to see how many batches were used
        )

        self.model.eval()
        validation_loss_sum = 0.0
        validation_metrics_sum = defaultdict(float)
        batch_counter = 0
        for batch_idx, val_batch in progress_bar:
            loss, metrics = self.validation_step(batch=val_batch, batch_idx=batch_idx)
            validation_loss_sum += loss
            batch_counter += 1
            for name, val in metrics.items():
                validation_metrics_sum[name] += val

            # Important: update counter
            self.validation_glob_step += 1

        # Aggregate and log losses
        avg_loss = validation_loss_sum / batch_counter
        self.log(
            item=avg_loss.item(),
            identifier="validation_loss_epoch",
            kind="metric",
            step=self.validation_glob_step,
        )
        # Aggregate and log metrics
        for m_name, m_val in validation_metrics_sum.items():
            self.log(
                item=m_val / batch_counter,
                identifier="validation_" + m_name + "_epoch",
                kind="metric",
                step=self.validation_glob_step,
            )

        return avg_loss

    def validation_step(
        self, batch: Batch, batch_idx: int
    ) -> Tuple[torch.Tensor, Dict[str, Any]]:
        """Perform a single optimization step using a batch sampled from the
        validation dataset.

        Args:
            batch (Batch): batch sampled by a dataloader.
            batch_idx (int): batch index in the dataloader.

        Returns:
            Tuple[Loss, Dict[str, Any]]: batch loss and dictionary of metric
            values with the same structure of ``self.metrics``.
        """
        x, y = batch
        x, y = x.to(self.device), y.to(self.device)
        with torch.no_grad():
            pred_y = self.model(x)
            loss: torch.Tensor = self.loss(pred_y, y)
        self.log(
            item=loss.item(),
            identifier="validation_loss",
            kind="metric",
            step=self.validation_glob_step,
            batch_idx=batch_idx,
        )
        metrics: Dict[str, Any] = self.compute_metrics(
            true=y,
            pred=pred_y,
            logger_step=self.validation_glob_step,
            batch_idx=batch_idx,
            stage="validation",
        )
        return loss, metrics

    def test_epoch(self, epoch: int) -> torch.Tensor:
        """Perform a complete sweep over the test dataset, completing an
        epoch of test.

        Args:
            epoch (int): current epoch number, from 0 to ``self.epochs - 1``.

        Returns:
            Loss: average test loss for the current epoch.
        """
        raise NotImplementedError()

    def test_step(self, batch: Batch, batch_idx: int) -> Tuple[torch.Tensor, Dict[str, Any]]:
        """Perform a single predictions step using a batch sampled from the
        test dataset.

        Args:
            batch (Batch): batch sampled by a dataloader.
            batch_idx (int): batch index in the dataloader.

        Returns:
            Tuple[Loss, Dict[str, Any]]: batch loss and dictionary of metric
            values with the same structure of ``self.metrics``.
        """
        raise NotImplementedError()


class GANTrainer(TorchTrainer):
    """Trainer class for GAN models using pytorch.

    Args:
        config (Union[Dict, TrainingConfiguration]): training configuration
            containing hyperparameters.
        epochs (int): number of training epochs.
        discriminator (nn.Module): pytorch discriminator model to train GAN.
        generator (nn.Module): pytorch generator model to train GAN.
        strategy (Literal['ddp', 'deepspeed', 'horovod'], optional):
            distributed strategy. Defaults to 'ddp'.
        validation_every (Optional[int], optional): run a validation epoch
            every ``validation_every`` epochs. Disabled if None. Defaults to 1.
        test_every (Optional[int], optional): run a test epoch
            every ``test_every`` epochs. Disabled if None. Defaults to None.
        random_seed (Optional[int], optional): set random seed for
            reproducibility. If None, the seed is not set. Defaults to None.
        logger (Optional[Logger], optional): logger for ML tracking.
            Defaults to None.
        metrics (Optional[Dict[str, Metric]], optional): map of torch metrics
            metrics. Defaults to None.
        checkpoints_location (str): path to checkpoints directory.
            Defaults to "checkpoints".
        checkpoint_every (Optional[int]): save a checkpoint every
            ``checkpoint_every`` epochs. Disabled if None. Defaults to None.
        name (Optional[str], optional): trainer custom name. Defaults to None.
    """

    def __init__(
        self,
        config: Union[Dict, TrainingConfiguration],
        epochs: int,
        discriminator: nn.Module,
        generator: nn.Module,
        strategy: Literal["ddp", "deepspeed"] = "ddp",
        validation_every: Optional[int] = 1,
        test_every: Optional[int] = None,
        random_seed: Optional[int] = None,
        logger: Optional[Logger] = None,
        metrics: Optional[Dict[str, Metric]] = None,
        checkpoints_location: str = "checkpoints",
        checkpoint_every: Optional[int] = None,
        name: Optional[str] = None,
        **kwargs,
    ) -> None:
        super().__init__(
            config=config,
            epochs=epochs,
            model=None,
            strategy=strategy,
            validation_every=validation_every,
            test_every=test_every,
            random_seed=random_seed,
            logger=logger,
            metrics=metrics,
            checkpoints_location=checkpoints_location,
            checkpoint_every=checkpoint_every,
            name=name,
            **kwargs,
        )
        self.save_parameters(**self.locals2params(locals()))
        self.discriminator = discriminator
        self.generator = generator

    def create_model_loss_optimizer(self) -> None:
        self.optimizerD = optim.Adam(
            self.discriminator.parameters(), lr=self.config.lr, betas=(0.5, 0.999)
        )
        self.optimizerG = optim.Adam(
            self.generator.parameters(), lr=self.config.lr, betas=(0.5, 0.999)
        )
        self.criterion = nn.BCELoss()

        # https://stackoverflow.com/a/67437077
        self.discriminator = torch.nn.SyncBatchNorm.convert_sync_batchnorm(self.discriminator)
        self.generator = torch.nn.SyncBatchNorm.convert_sync_batchnorm(self.generator)

        # First, define strategy-wise optional configurations
        if isinstance(self.strategy, DeepSpeedStrategy):
            # Batch size definition is not optional for DeepSpeedStrategy!
            distribute_kwargs = dict(
                config_params=dict(train_micro_batch_size_per_gpu=self.config.batch_size)
            )
        else:
            distribute_kwargs = {}
        # Distribute discriminator and its optimizer
        self.discriminator, self.optimizerD, _ = self.strategy.distributed(
            self.discriminator, self.optimizerD, **distribute_kwargs
        )
        self.generator, self.optimizerG, _ = self.strategy.distributed(
            self.generator, self.optimizerG, **distribute_kwargs
        )

    def train_epoch(self, epoch: int):
        self.discriminator.train()
        self.generator.train()
        gen_train_losses = []
        disc_train_losses = []
        disc_train_accuracy = []
        for batch_idx, (real_images, _) in enumerate(self.train_dataloader):
            lossG, lossD, accuracy_disc = self.train_step(real_images, batch_idx)
            gen_train_losses.append(lossG)
            disc_train_losses.append(lossD)
            disc_train_accuracy.append(accuracy_disc)

            self.train_glob_step += 1
        # Aggregate and log losses and accuracy
        avg_disc_accuracy = torch.mean(torch.stack(disc_train_accuracy))
        self.log(
            item=avg_disc_accuracy.item(),
            identifier="disc_train_accuracy_per_epoch",
            kind="metric",
            step=epoch,
        )
        avg_gen_loss = torch.mean(torch.stack(gen_train_losses))
        self.log(
            item=avg_gen_loss.item(),
            identifier="gen_train_loss_per_epoch",
            kind="metric",
            step=epoch,
        )

        avg_disc_loss = torch.mean(torch.stack(disc_train_losses))
        self.log(
            item=avg_disc_loss.item(),
            identifier="disc_train_loss_per_epoch",
            kind="metric",
            step=epoch,
        )

        self.save_fake_generator_images(epoch)

    def validation_epoch(self, epoch: int):
        gen_validation_losses = []
        gen_validation_accuracy = []
        disc_validation_losses = []
        disc_validation_accuracy = []
        self.discriminator.eval()
        self.generator.eval()
        for batch_idx, (real_images, _) in enumerate(self.validation_dataloader):
            loss_gen, accuracy_gen, loss_disc, accuracy_disc = self.validation_step(
                real_images, batch_idx
            )
            gen_validation_losses.append(loss_gen)
            gen_validation_accuracy.append(accuracy_gen)
            disc_validation_losses.append(loss_disc)
            disc_validation_accuracy.append(accuracy_disc)
            self.validation_glob_step += 1

        # Aggregate and log metrics
        disc_validation_loss = torch.mean(torch.stack(disc_validation_losses))
        self.log(
            item=disc_validation_loss.item(),
            identifier="disc_valid_loss_per_epoch",
            kind="metric",
            step=epoch,
        )
        disc_validation_accuracy = torch.mean(torch.stack(disc_validation_accuracy))
        self.log(
            item=disc_validation_accuracy.item(),
            identifier="disc_valid_accuracy_epoch",
            kind="metric",
            step=epoch,
        )
        gen_validation_loss = torch.mean(torch.stack(gen_validation_losses))
        self.log(
            item=gen_validation_loss.item(),
            identifier="gen_valid_loss_per_epoch",
            kind="metric",
            step=epoch,
        )
        gen_validation_accuracy = torch.mean(torch.stack(gen_validation_accuracy))
        self.log(
            item=gen_validation_accuracy.item(),
            identifier="gen_valid_accuracy_epoch",
            kind="metric",
            step=epoch,
        )

        return gen_validation_loss

    def train_step(self, real_images, batch_idx):
        real_images = real_images.to(self.device)
        batch_size = real_images.size(0)
        real_labels = torch.ones((batch_size,), dtype=torch.float, device=self.device)
        fake_labels = torch.zeros((batch_size,), dtype=torch.float, device=self.device)

        # Train Discriminator with real images
        output_real = self.discriminator(real_images)
        lossD_real = self.criterion(output_real, real_labels)
        # Generate fake images and train Discriminator
        noise = torch.randn(batch_size, self.config.z_dim, 1, 1, device=self.device)

        fake_images = self.generator(noise)
        output_fake = self.discriminator(fake_images.detach())
        lossD_fake = self.criterion(output_fake, fake_labels)

        lossD = (lossD_real + lossD_fake) / 2

        self.optimizerD.zero_grad()
        lossD.backward()
        self.optimizerD.step()

        accuracy = ((output_real > 0.5).float() == real_labels).float().mean() + (
            (output_fake < 0.5).float() == fake_labels
        ).float().mean()
        accuracy_disc = accuracy.mean()

        # Train Generator
        output_fake = self.discriminator(fake_images)
        lossG = self.criterion(output_fake, real_labels)
        self.optimizerG.zero_grad()
        lossG.backward()
        self.optimizerG.step()
        self.log(
            item=accuracy_disc,
            identifier="disc_train_accuracy_per_batch",
            kind="metric",
            step=self.train_glob_step,
            batch_idx=batch_idx,
        )
        self.log(
            item=lossG,
            identifier="gen_train_loss_per_batch",
            kind="metric",
            step=self.train_glob_step,
            batch_idx=batch_idx,
        )
        self.log(
            item=lossD,
            identifier="disc_train_loss_per_batch",
            kind="metric",
            step=self.train_glob_step,
            batch_idx=batch_idx,
        )

        return lossG, lossD, accuracy_disc

    def validation_step(self, real_images, batch_idx):
        real_images = real_images.to(self.device)
        batch_size = real_images.size(0)
        real_labels = torch.ones((batch_size,), dtype=torch.float, device=self.device)
        fake_labels = torch.zeros((batch_size,), dtype=torch.float, device=self.device)

        # Validate with real images
        output_real = self.discriminator(real_images)
        loss_real = self.criterion(output_real, real_labels)

        # Generate and validate fake images
        noise = torch.randn(batch_size, self.config.z_dim, 1, 1, device=self.device)

        with torch.no_grad():
            fake_images = self.generator(noise)
            output_fake = self.discriminator(fake_images.detach())
        loss_fake = self.criterion(output_fake, fake_labels)

        # Generator's attempt to fool the discriminator
        loss_gen = self.criterion(output_fake, real_labels)
        accuracy_gen = ((output_fake > 0.5).float() == real_labels).float().mean()

        # Calculate total discriminator loss and accuracy
        d_total_loss = (loss_real + loss_fake) / 2
        accuracy = ((output_real > 0.5).float() == real_labels).float().mean() + (
            (output_fake < 0.5).float() == fake_labels
        ).float().mean()
        d_accuracy = accuracy.item() / 2

        self.log(
            item=loss_gen.item(),
            identifier="gen_valid_loss_per_batch",
            kind="metric",
            step=self.validation_glob_step,
            batch_idx=batch_idx,
        )
        self.log(
            item=accuracy_gen.item(),
            identifier="gen_valid_accuracy_per_batch",
            kind="metric",
            step=self.validation_glob_step,
            batch_idx=batch_idx,
        )

        self.log(
            item=d_total_loss.item(),
            identifier="disc_valid_loss_per_batch",
            kind="metric",
            step=self.validation_glob_step,
            batch_idx=batch_idx,
        )
        self.log(
            item=d_accuracy,
            identifier="disc_valid_accuracy_per_batch",
            kind="metric",
            step=self.validation_glob_step,
            batch_idx=batch_idx,
        )
        return loss_gen, accuracy_gen

    def save_checkpoint(self, name, epoch, loss=None):
        """Save training checkpoint with both optimizers."""
        if not os.path.exists(self.checkpoints_location):
            os.makedirs(self.checkpoints_location)

        checkpoint_path = os.path.join(self.checkpoints_location, f"{name}")
        checkpoint = {
            "epoch": epoch,
            "loss": loss.item() if loss is not None else None,
            "discriminator_state_dict": self.discriminator.state_dict(),
            "generator_state_dict": self.generator.state_dict(),
            "optimizerD_state_dict": self.optimizerD.state_dict(),
            "optimizerG_state_dict": self.optimizerG.state_dict(),
            "lr_scheduler": self.lr_scheduler.state_dict() if self.lr_scheduler else None,
        }

        torch.save(checkpoint, checkpoint_path)
        print(f"Checkpoint saved to {checkpoint_path}")

    def load_checkpoint(self, checkpoint_path):
        """Load models and optimizers from checkpoint."""
        checkpoint = torch.load(checkpoint_path)

        self.discriminator.load_state_dict(checkpoint["discriminator_state_dict"])
        self.generator.load_state_dict(checkpoint["generator_state_dict"])
        self.optimizerD.load_state_dict(checkpoint["optimizerD_state_dict"])
        self.optimizerG.load_state_dict(checkpoint["optimizerG_state_dict"])

        if "lr_scheduler" in checkpoint:
            if checkpoint["lr_scheduler"] is not None:
                self.lr_scheduler.load_state_dict(checkpoint["lr_scheduler"])

        print(f"Checkpoint loaded from {checkpoint_path}")

    def save_fake_generator_images(self, epoch):
        """
        plot and save fake images from generator

         Args:
            epoch (int): epoch number, from 0 to ``epochs-1``.
        """
        import matplotlib.pyplot as plt
        import numpy as np

        self.generator.eval()
        noise = torch.randn(64, self.config.z_dim, 1, 1, device=self.device)
        fake_images = self.generator(noise)
        fake_images_grid = torchvision.utils.make_grid(fake_images, normalize=True)
        fig, ax = plt.subplots(figsize=(8, 8))
        ax.set_axis_off()
        ax.set_title(f"Fake images for epoch {epoch}")
        ax.imshow(np.transpose(fake_images_grid.cpu().numpy(), (1, 2, 0)))
        self.log(
            item=fig,
            identifier=f"fake_images_epoch_{epoch}.png",
            kind="figure",
            step=epoch,
        )


class TorchLightningTrainer(Trainer):
    """Generic trainer for torch Lightning workflows.

    Args:
        config (Union[Dict, str]): `Lightning configuration`_
            which can be the path to a file or a Python dictionary.
        mlflow_saved_model (str, optional): name of the model created in
            MLFlow. Defaults to 'my_model'.

    .. _Lightning configuration:
        https://pytorch-lightning.readthedocs.io/en/1.6.5/common/lightning_cli.html
    """

    def __init__(self, config: Union[Dict, str], mlflow_saved_model: str = "my_model"):
        self.save_parameters(**self.locals2params(locals()))
        super().__init__()
        if isinstance(config, str) and os.path.isfile(config):
            # Load from YAML
            config = load_yaml(config)
        self.conf = config
        self.mlflow_saved_model = mlflow_saved_model

    @monitor_exec
    def execute(self) -> Any:
        import lightning as L
        from lightning.pytorch.cli import LightningCLI

        from .mlflow import init_lightning_mlflow, teardown_lightning_mlflow

        init_lightning_mlflow(
            self.conf, tmp_dir="/tmp", registered_model_name=self.mlflow_saved_model
        )
        old_argv = sys.argv
        sys.argv = ["some_script_placeholder.py"]
        cli = LightningCLI(
            args=self.conf,
            model_class=L.LightningModule,
            datamodule_class=L.LightningDataModule,
            run=False,
            save_config_kwargs={
                "overwrite": True,
                "config_filename": "pl-training.yml",
            },
            subclass_mode_model=True,
            subclass_mode_data=True,
        )
        sys.argv = old_argv
        cli.trainer.fit(cli.model, datamodule=cli.datamodule)
        teardown_lightning_mlflow()


def _distributed_dataloader(dataloader: DataLoader, gwsize, grank):
    """Makes a Dataloader distributed."""
    sampler = DistributedSampler(
        dataloader.dataset, num_replicas=gwsize, rank=grank, shuffle=True
    )
    # Recreate dataloader, with updated sampler
    return DataLoader(
        dataloader.dataset,
        batch_size=dataloader.batch_size,
        sampler=sampler,
        num_workers=dataloader.num_workers,
        collate_fn=dataloader.collate_fn,
        pin_memory=dataloader.pin_memory,
        drop_last=dataloader.drop_last,
        timeout=dataloader.timeout,
        worker_init_fn=seed_worker,  # dataloader.worker_init_fn,
        multiprocessing_context=dataloader.multiprocessing_context,
        generator=dataloader.generator,
        prefetch_factor=dataloader.prefetch_factor,
        persistent_workers=dataloader.persistent_workers,
        pin_memory_device=dataloader.pin_memory_device,
    )


def distributed(func):
    """The decorated function must have a standard signature.
    Its first arguments must be:
    model, train_dataloader, validation_dataloader, device (in this order).

    Additional args or kwargs are allowed consistently with the signature
    of the decorated function.
    """

    def dist_train(
        model, train_dataloader, validation_dataloader=None, device="cpu", *args, **kwargs
    ):
        if torch.cuda.is_available():
            dist.init_process_group(backend="nccl")

        if torch.cuda.is_available():
            lwsize = torch.cuda.device_count()  # local world size - per node
            gwsize = dist.get_world_size()  # global world size - per run
            grank = dist.get_rank()  # global rank - assign per run
            lrank = dist.get_rank() % lwsize  # local rank - assign per node
        else:
            gwsize = 1
            grank = 0
            lrank = 0

        device = torch.device("cuda" if torch.cuda.is_available() else "cpu", lrank)
        if torch.cuda.is_available():
            torch.cuda.set_device(lrank)

        model = model.to(device)
        model = DDP(model, device_ids=[device], output_device=device)

        train_dataloader = _distributed_dataloader(train_dataloader, gwsize, grank)
        if validation_dataloader is not None:
            validation_dataloader = _distributed_dataloader(
                validation_dataloader, gwsize, grank
            )

        try:
            func(model, train_dataloader, validation_dataloader, device, *args, **kwargs)
        finally:
            if torch.cuda.is_available():
                dist.barrier()
                dist.destroy_process_group()

    return dist_train


class RayTorchTrainer(Trainer):
    """A trainer class for distributed training and hyperparameter optimization
    using Ray Train/ Tune and PyTorch.

    Args:
        config (Dict): A dictionary of configuration settings for the trainer.
        strategy (Optional[Literal["ddp", "deepspeed"]]):
            The distributed training strategy to use. Defaults to "ddp".
        name (Optional[str]): Optional name for the trainer instance. Defaults to None.
        logger (Optional[Logger]): Optional logger instance. Defaults to None.
    """

    def __init__(
        self,
        config: Dict,
        strategy: Literal["ddp", "deepspeed"] = "ddp",
        name: str | None = None,
        logger: Logger | None = None,
        random_seed: int = 1234,
    ) -> None:
        super().__init__(name=name)
        self.logger = logger
        self._set_strategy_and_init_ray(strategy)
        self._set_configs(config=config)
        self.torch_rng = set_seed(random_seed)

        import ray.train
        import ray.tune

        self.ray_train = ray.train
        self.ray_tune = ray.tune

    def _set_strategy_and_init_ray(self, strategy: str) -> None:
        """Set the distributed training strategy. This will initialize the ray backend.

        Args:
            strategy (str): The strategy to use for distributed training.
                Must be one of ["ddp", "deepspeed"].

        Raises:
            ValueError: If an unsupported strategy is provided.
        """
        if strategy == "ddp":
            self.strategy = RayDDPStrategy()
        elif strategy == "deepspeed":
            self.strategy = RayDeepSpeedStrategy(backend="nccl")
        else:
            raise ValueError(f"Unsupported strategy: {strategy}")

    def _set_configs(self, config: Dict) -> None:
        self.config = config
        self._set_scaling_config()
        self._set_tune_config()
        self._set_run_config()
        self._set_train_loop_config()

    @property
    def device(self) -> str:
        """Get the current device from distributed strategy.
        Returns:
            str: Device string (e.g., "cuda:0").
        """
        return self.strategy.device()

    def create_dataloaders(
        self,
        train_dataset: Dataset,
        validation_dataset: Dataset | None = None,
        test_dataset: Dataset | None = None,
        batch_size: int = 1,
        num_workers_dataloader: int = 4,
        pin_memory: bool = False,
        shuffle_train: bool | None = False,
        shuffle_test: bool | None = False,
        shuffle_validation: bool | None = False,
        sampler: Union[Sampler, Iterable, None] = None,
        collate_fn: Callable[[List], Any] | None = None,
    ) -> None:
        """Create data loaders for training, validation, and testing.

        Args:
            train_dataset (Dataset): The training dataset.
            validation_dataset (Dataset, optional): The validation dataset. Defaults to None.
            test_dataset (Dataset, optional): The test dataset. Defaults to None.
            batch_size (int, optional): Batch size for data loaders. Defaults to 1.
            shuffle_train (bool, optional): Whether to shuffle the training dataset.
                Defaults to False.
            shuffle_test (bool, optional): Whether to shuffle the test dataset.
                Defaults to False.
            shuffle_validation (bool, optional): Whether to shuffle the validation dataset.
                Defaults to False.
            sampler (Union[Sampler, Iterable, None], optional): Sampler for the datasets.
                Defaults to None.
            collate_fn (Callable[[List], Any], optional):
                Function to collate data samples into batches. Defaults to None.
        """
        self.train_dataloader = self.strategy.create_dataloader(
            dataset=train_dataset,
            batch_size=batch_size,
            num_workers=num_workers_dataloader,
            pin_memory=pin_memory,
            generator=self.torch_rng,
            shuffle=shuffle_train,
            sampler=sampler,
            collate_fn=collate_fn,
        )
        if validation_dataset is not None:
            self.validation_dataloader = self.strategy.create_dataloader(
                dataset=validation_dataset,
                batch_size=batch_size,
                num_workers=num_workers_dataloader,
                pin_memory=pin_memory,
                generator=self.torch_rng,
                shuffle=shuffle_validation,
                sampler=sampler,
                collate_fn=collate_fn,
            )
        else:
            self.validation_dataloader = None
        if test_dataset is not None:
            self.test_dataloader = self.strategy.create_dataloader(
                dataset=test_dataset,
                batch_size=batch_size,
                num_workers=num_workers_dataloader,
                pin_memory=pin_memory,
                generator=self.torch_rng,
                shuffle=shuffle_test,
                sampler=sampler,
                collate_fn=collate_fn,
            )
        else:
            self.test_dataloader = None

    @monitor_exec
    def execute(
        self,
        train_dataset: Dataset,
        validation_dataset: Dataset | None = None,
        test_dataset: Dataset | None = None,
    ) -> Tuple[Dataset, Dataset, Dataset, Any]:
        """Execute the training pipeline with the given datasets.

        Args:
            train_dataset (Dataset): Training dataset.
            validation_dataset (Optional[Dataset], optional): Validation dataset.
                Defaults to None.
            test_dataset (Optional[Dataset], optional): Test dataset. Defaults to None.

        Returns:
            Tuple[Dataset, Dataset, Dataset, Any]:
                A tuple containing the datasets and the training result grid.
        """
        import ray.train.torch

        train_with_data = self.ray_tune.with_parameters(
            self.train, data=[train_dataset, validation_dataset, test_dataset]
        )
        trainer = ray.train.torch.TorchTrainer(
            train_with_data,
            train_loop_config=self.train_loop_config,
            scaling_config=self.scaling_config,
            run_config=self.run_config,
        )
        param_space = {"train_loop_config": self.train_loop_config}
        tuner = self.ray_tune.Tuner(
            trainer, param_space=param_space, tune_config=self.tune_config
        )

        result_grid = tuner.fit()

        return train_dataset, validation_dataset, test_dataset, result_grid

    def set_epoch(self, epoch: int) -> None:
        self.train_dataloader.sampler.set_epoch(epoch)
        if self.validation_dataloader is not None:
            self.validation_dataloader.sampler.set_epoch(epoch)
        if self.test_dataloader is not None:
            self.test_dataloader.sampler.set_epoch(epoch)

    def _set_tune_config(self) -> None:
        from .tuning import get_raytune_scheduler, get_raytune_search_alg

        tune_config = self.config.get("tune_config", {})

        if not tune_config:
            print(
                "WARNING: Empty Tune Config configured. Using the default configuration with "
                "a single trial."
            )

        search_alg = get_raytune_search_alg(tune_config)
        scheduler = get_raytune_scheduler(tune_config)

        metric = tune_config.get("metric", "loss")
        mode = tune_config.get("mode", "min")

        try:
            self.tune_config = self.ray_tune.TuneConfig(
                **tune_config,
                search_alg=search_alg,
                scheduler=scheduler,
                metric=metric,
                mode=mode,
            )
        except AttributeError as e:
            print(
                "Could not set Tune Config. Please ensure that you have passed the "
                "correct arguments for it. You can find more information for which "
                "arguments to set at "
                "https://docs.ray.io/en/latest/tune/api/doc/ray.tune.TuneConfig.html."
            )
            print(e)

    def _set_scaling_config(self) -> None:
        scaling_config = self.config.get("scaling_config", {})

        if not scaling_config:
            print("WARNING: No Scaling Config configured. Running trials non-distributed.")

        try:
            self.scaling_config = self.ray_train.ScalingConfig(**scaling_config)
        except AttributeError as e:
            print(
                "Could not set Scaling Config. Please ensure that you have passed the "
                "correct arguments for it. You can find more information for which "
                "arguments to set at "
                "https://docs.ray.io/en/latest/train/api/doc/ray.train.ScalingConfig.html"
            )
            print(e)

    def _set_run_config(self) -> None:
        run_config = self.config.get("run_config", {})

        if not run_config:
            print("WARNING: No RunConfig provided. Assuming local or single-node execution.")

        try:
            storage_path = Path(run_config.pop("storage_path")).resolve()

            if not storage_path:
                print(
                    "INFO: Empty storage path provided. Using default path 'ray_checkpoints'"
                )
                storage_path = Path("ray_checkpoints").resolve()

            self.run_config = self.ray_train.RunConfig(**run_config, storage_path=storage_path)
        except AttributeError as e:
            print(
                "Could not set Run Config. Please ensure that you have passed the "
                "correct arguments for it. You can find more information for which "
                "arguments to set at "
                "https://docs.ray.io/en/latest/train/api/doc/ray.train.RunConfig.html"
            )
            print(e)

    def _set_train_loop_config(self) -> None:
        self.train_loop_config = self.config.get("train_loop_config", {})

        if not self.train_loop_config:
            print(
                "WARNING: No training_loop_config detected. "
                "If you want to tune any hyperparameters, make sure to define them here."
            )
            return

        try:
            for name, param in self.train_loop_config.items():
                if not isinstance(param, dict):
                    continue

                # Convert specific keys to float if necessary
                for key in ["lower", "upper", "mean", "std"]:
                    if key in param:
                        param[key] = float(param[key])

                param_type = param.pop("type")
                param = getattr(self.ray_tune, param_type)(**param)
                self.train_loop_config[name] = param

        except AttributeError as e:
            print(
                f"{param} could not be set. Check that this parameter type is "
                "supported by Ray Tune at "
                "https://docs.ray.io/en/latest/tune/api/search_space.html"
            )
            print(e)

    # TODO: Can I also log the checkpoint?
    def checkpoint_and_report(self, epoch, tuning_metrics, checkpointing_data=None):
        with tempfile.TemporaryDirectory() as temp_checkpoint_dir:
            checkpoint = None

            should_checkpoint = epoch % self.config.get("checkpoint_freq", 1)

            if checkpointing_data and should_checkpoint:
                torch.save(checkpointing_data, os.path.join(temp_checkpoint_dir, str(epoch)))
                checkpoint = self.ray_train.Checkpoint.from_directory(temp_checkpoint_dir)

        self.ray_train.report(tuning_metrics, checkpoint=checkpoint)

    def initialize_logger(self, hyperparams: Optional[Dict], rank):
        if not self.logger:
            return

        self.logger.create_logger_context(rank=rank)
        print(f"Logger initialized with rank {rank}")

        if hyperparams:
            self.logger.save_hyperparameters(hyperparams)

    def close_logger(self):
        if self.logger:
            self.logger.destroy_logger_context()

    def log(
        self,
        item: Union[Any, List[Any]],
        identifier: Union[str, List[str]],
        kind: str = "metric",
        step: Optional[int] = None,
        batch_idx: Optional[int] = None,
        **kwargs,
    ) -> None:
        """Log ``item`` with ``identifier`` name of ``kind`` type at ``step``
        time step.

        Args:
            item (Union[Any, List[Any]]): element to be logged (e.g., metric).
            identifier (Union[str, List[str]]): unique identifier for the
                element to log(e.g., name of a metric).
            kind (str, optional): type of the item to be logged. Must be one
                among the list of self.supported_types. Defaults to 'metric'.
            step (Optional[int], optional): logging step. Defaults to None.
            batch_idx (Optional[int], optional): DataLoader batch counter
                (i.e., batch idx), if available. Defaults to None.
        """
        if self.logger:
            self.logger.log(
                item=item,
                identifier=identifier,
                kind=kind,
                step=step,
                batch_idx=batch_idx,
                **kwargs,
            )
        else:
            print(
                "INFO: The log method was called, but no logger was configured for this "
                "Trainer."
            )



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

import logging
import os
from unittest.mock import MagicMock, patch

import pytest
import torch.nn.functional as F
from torch import nn
from torchvision import datasets, transforms

from itwinai.torch.trainer import TorchTrainer

MNIST_PATH = "mnist_dataset"


class Net(nn.Module):
    def __init__(self):
        super(Net, self).__init__()
        self.conv1 = nn.Conv2d(1, 10, kernel_size=5)
        self.conv2 = nn.Conv2d(10, 20, kernel_size=5)
        self.conv2_drop = nn.Dropout2d()
        self.fc1 = nn.Linear(320, 50)
        self.fc2 = nn.Linear(50, 10)

    def forward(self, x):
        x = F.relu(F.max_pool2d(self.conv1(x), 2))
        x = F.relu(F.max_pool2d(self.conv2_drop(self.conv2(x)), 2))
        x = x.view(-1, 320)
        x = F.relu(self.fc1(x))
        x = F.dropout(x, training=self.training)
        x = self.fc2(x)
        return F.log_softmax(x, dim=0)


@pytest.fixture(scope="module")
def mnist_datasets():
    """Parse MNIST datasets."""
    if not os.environ.get("MNIST_PATH"):
        logging.warning("MNIST dataset not found locally. I have to download it!")

    dataset_path = os.environ.get("MNIST_PATH", MNIST_PATH)
    train_set = datasets.MNIST(
        dataset_path,
        train=True,
        download=True,
        transform=transforms.Compose(
            [transforms.ToTensor(), transforms.Normalize((0.1307,), (0.3081,))]
        ),
    )
    val_set = datasets.MNIST(
        dataset_path,
        train=False,
        download=False,
        transform=transforms.Compose(
            [transforms.ToTensor(), transforms.Normalize((0.1307,), (0.3081,))]
        ),
    )
    return train_set, val_set


@pytest.mark.hpc
@pytest.mark.parametrize(
    "strategy_name,strategy_fixture",
    [
        pytest.param("ddp", "ddp_strategy", marks=pytest.mark.torch_dist),
        pytest.param("deepspeed", "deepspeed_strategy", marks=pytest.mark.deepspeed_dist),
        pytest.param("horovod", "horovod_strategy", marks=pytest.mark.horovod_dist),
    ],
)
def test_distributed_trainer_mnist(mnist_datasets, request, strategy_name, strategy_fixture):
    """Test TorchTrainer on MNIST with different distributed strategies."""
    training_config = dict(optimizer="sgd", loss="nllloss")
    trainer = TorchTrainer(
        model=Net(),
        config=training_config,
        epochs=2,
        strategy=strategy_name,
        checkpoint_every=1,
    )

    strategy_instance = request.getfixturevalue(strategy_fixture)
    trainer.strategy = strategy_instance  # Patch the strategy with the fixture instance

    train_set, val_set = mnist_datasets

    # Mock strategy cleanup
    with patch.object(
        trainer.strategy, "clean_up", new=MagicMock(name="clean_up")
    ) as mock_cleanup:
        trainer.execute(train_set, val_set)

        # Check that the torch trainer is cleaning up the strategy
        mock_cleanup.assert_called_once()
